"""String operations."""

from __future__ import annotations

from typing import Optional

from public import public

import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.rules as rlz
from xorq.vendor.ibis.common.annotations import attribute
from xorq.vendor.ibis.common.typing import VarTuple  # noqa: TCH001
from xorq.vendor.ibis.expr.operations.core import Unary, Value


@public
class StringUnary(Unary):
    """Base class for string operations accepting one argument."""

    arg: Value[dt.String]

    dtype = dt.string


@public
class Uppercase(StringUnary):
    """Convert a string to uppercase."""


@public
class Lowercase(StringUnary):
    """Convert a string to lowercase."""


@public
class Reverse(StringUnary):
    """Reverse a string."""


@public
class Strip(StringUnary):
    """Strip leading and trailing whitespace."""


@public
class LStrip(StringUnary):
    """Strip leading whitespace."""


@public
class RStrip(StringUnary):
    """Strip trailing whitespace."""


@public
class Capitalize(StringUnary):
    """Capitalize the first letter of a string."""


@public
class Substring(Value):
    """Extract a substring from a string."""

    arg: Value[dt.String]
    start: Value[dt.Integer]
    length: Optional[Value[dt.Integer]] = None

    dtype = dt.string
    shape = rlz.shape_like("args")


@public
class StringSlice(Value):
    """Extract a substring from a string."""

    arg: Value[dt.String]
    start: Optional[Value[dt.Integer]] = None
    end: Optional[Value[dt.Integer]] = None

    dtype = dt.string
    shape = rlz.shape_like("args")


@public
class StrRight(Value):
    """Extract a substring starting from the right of a string."""

    arg: Value[dt.String]
    nchars: Value[dt.Integer]

    shape = rlz.shape_like("args")
    dtype = dt.string


@public
class Repeat(Value):
    """Repeat a string."""

    arg: Value[dt.String]
    times: Value[dt.Integer]

    shape = rlz.shape_like("args")
    dtype = dt.string


@public
class StringFind(Value):
    """Find the position of a substring in a string."""

    arg: Value[dt.String]
    substr: Value[dt.String]
    start: Optional[Value[dt.Integer]] = None
    end: Optional[Value[dt.Integer]] = None

    shape = rlz.shape_like("args")
    dtype = dt.int64


@public
class Translate(Value):
    """Translate characters in a string."""

    arg: Value[dt.String]
    from_str: Value[dt.String]
    to_str: Value[dt.String]

    shape = rlz.shape_like("args")
    dtype = dt.string


@public
class LPad(Value):
    """Pad a string on the left."""

    arg: Value[dt.String]
    length: Value[dt.Integer]
    pad: Optional[Value[dt.String]] = None

    shape = rlz.shape_like("args")
    dtype = dt.string


@public
class RPad(Value):
    """Pad a string on the right."""

    arg: Value[dt.String]
    length: Value[dt.Integer]
    pad: Optional[Value[dt.String]] = None

    shape = rlz.shape_like("args")
    dtype = dt.string


@public
class FindInSet(Value):
    """Find the position of a string in a list of comma-separated strings."""

    needle: Value[dt.String]
    values: VarTuple[Value[dt.String]]

    shape = rlz.shape_like("needle")
    dtype = dt.int64


@public
class StringJoin(Value):
    """Join strings with a separator."""

    arg: VarTuple[Value[dt.String]]
    sep: Value[dt.String]

    dtype = dt.string

    @attribute
    def shape(self):
        return rlz.highest_precedence_shape((self.sep, *self.arg))


@public
class ArrayStringJoin(Value):
    """Join strings in an array with a separator."""

    arg: Value[dt.Array[dt.String]]
    sep: Value[dt.String]

    dtype = dt.string
    shape = rlz.shape_like("args")


@public
class StartsWith(Value):
    """Check if a string starts with another string."""

    arg: Value[dt.String]
    start: Value[dt.String]

    dtype = dt.boolean
    shape = rlz.shape_like("args")


@public
class EndsWith(Value):
    """Check if a string ends with another string."""

    arg: Value[dt.String]
    end: Value[dt.String]

    dtype = dt.boolean
    shape = rlz.shape_like("args")


@public
class FuzzySearch(Value):
    arg: Value[dt.String]
    pattern: Value[dt.String]

    dtype = dt.boolean
    shape = rlz.shape_like("args")


@public
class StringSQLLike(FuzzySearch):
    """SQL LIKE string match operation.

    Similar to globbing.
    """

    arg: Value[dt.String]
    pattern: Value[dt.String]
    escape: Optional[str] = None


@public
class StringSQLILike(StringSQLLike):
    """Case-insensitive SQL LIKE string match operation.

    Similar to case-insensitive globbing.
    """


@public
class RegexSearch(FuzzySearch):
    """Search a string with a regular expression."""


@public
class RegexExtract(Value):
    """Extract a substring from a string using a regular expression."""

    arg: Value[dt.String]
    pattern: Value[dt.String]
    index: Value[dt.Integer]

    shape = rlz.shape_like("args")
    dtype = dt.string


@public
class RegexSplit(Value):
    """Split a string using a regular expression."""

    arg: Value[dt.String]
    pattern: Value[dt.String]

    shape = rlz.shape_like("args")
    dtype = dt.Array(dt.string)


@public
class RegexReplace(Value):
    """Replace a substring in a string using a regular expression."""

    arg: Value[dt.String]
    pattern: Value[dt.String]
    replacement: Value[dt.String]

    shape = rlz.shape_like("args")
    dtype = dt.string


@public
class StringReplace(Value):
    """Replace a substring in a string with another string."""

    arg: Value[dt.String]
    pattern: Value[dt.String]
    replacement: Value[dt.String]

    shape = rlz.shape_like("args")
    dtype = dt.string


@public
class StringSplit(Value):
    """Split a string using a delimiter."""

    arg: Value[dt.String]
    delimiter: Value[dt.String]

    shape = rlz.shape_like("args")
    dtype = dt.Array(dt.string)


@public
class StringConcat(Value):
    """Concatenate strings."""

    arg: VarTuple[Value[dt.String]]

    shape = rlz.shape_like("arg")
    dtype = rlz.dtype_like("arg")


@public
class ExtractURLField(StringUnary):
    pass


@public
class ExtractProtocol(ExtractURLField):
    """Extract the protocol from a URL."""


@public
class ExtractAuthority(ExtractURLField):
    """Extract the authority from a URL."""


@public
class ExtractUserInfo(ExtractURLField):
    """Extract the user info from a URL."""


@public
class ExtractHost(ExtractURLField):
    """Extract the host from a URL."""


@public
class ExtractFile(ExtractURLField):
    """Extract the file from a URL."""


@public
class ExtractPath(ExtractURLField):
    """Extract the path from a URL."""


@public
class ExtractQuery(ExtractURLField):
    """Extract the query from a URL."""

    key: Optional[Value[dt.String]] = None


@public
class ExtractFragment(ExtractURLField):
    """Extract the fragment from a URL."""


@public
class StringLength(StringUnary):
    """Compute the length of a string."""

    dtype = dt.int32


@public
class StringAscii(StringUnary):
    """Compute the ASCII code of the first character of a string."""

    dtype = dt.int32


@public
class StringContains(Value):
    """Check if a string contains a substring."""

    haystack: Value[dt.String]
    needle: Value[dt.String]

    shape = rlz.shape_like("args")
    dtype = dt.bool


@public
class Levenshtein(Value):
    """Compute the Levenshtein distance between two strings."""

    left: Value[dt.String]
    right: Value[dt.String]

    dtype = dt.int64
    shape = rlz.shape_like("args")



"""Operations for array expressions."""

from __future__ import annotations

from typing import Optional

from public import public

import xorq.vendor.ibis.expr.datashape as ds
import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.rules as rlz
from xorq.vendor.ibis.common.annotations import attribute
from xorq.vendor.ibis.common.typing import VarTuple  # noqa: TCH001
from xorq.vendor.ibis.expr.operations.core import Unary, Value


@public
class Array(Value):
    """Construct an array."""

    exprs: VarTuple[Value]

    @attribute
    def shape(self):
        return rlz.highest_precedence_shape(self.exprs)

    @attribute
    def dtype(self):
        return dt.Array(rlz.highest_precedence_dtype(self.exprs))


@public
class ArrayLength(Unary):
    """Compute the length of an array."""

    arg: Value[dt.Array]

    dtype = dt.int64
    shape = rlz.shape_like("args")


@public
class ArraySlice(Value):
    """Slice an array element."""

    arg: Value[dt.Array]
    start: Value[dt.Integer]
    stop: Optional[Value[dt.Integer]] = None

    dtype = rlz.dtype_like("arg")
    shape = rlz.shape_like("args")


@public
class ArrayIndex(Value):
    """Return the element of an array at some index."""

    arg: Value[dt.Array]
    index: Value[dt.Integer]

    shape = rlz.shape_like("args")

    @attribute
    def dtype(self):
        return self.arg.dtype.value_type


@public
class ArrayConcat(Value):
    """Concatenate two or more arrays into a single array."""

    arg: VarTuple[Value[dt.Array]]

    shape = rlz.shape_like("arg")

    @attribute
    def dtype(self):
        return dt.Array(dt.highest_precedence(arg.dtype.value_type for arg in self.arg))


@public
class ArrayRepeat(Value):
    """Repeat the elements of an array."""

    arg: Value[dt.Array]
    times: Value[dt.Integer]

    dtype = rlz.dtype_like("arg")
    shape = rlz.shape_like("args")


@public
class ArrayMap(Value):
    """Apply a function to every element of an array."""

    arg: Value[dt.Array]
    body: Value
    param: str

    shape = rlz.shape_like("arg")

    @attribute
    def dtype(self) -> dt.DataType:
        return dt.Array(self.body.dtype)


@public
class ArrayFilter(Value):
    """Filter array elements with a function."""

    arg: Value[dt.Array]
    body: Value[dt.Boolean]
    param: str

    shape = rlz.shape_like("arg")
    dtype = rlz.dtype_like("arg")


@public
class Unnest(Value):
    """Unnest an array value into a column."""

    arg: Value[dt.Array]

    shape = ds.columnar

    @attribute
    def dtype(self):
        return self.arg.dtype.value_type


@public
class ArrayContains(Value):
    """Return whether an array contains a specific value."""

    arg: Value[dt.Array]
    other: Value

    dtype = dt.boolean
    shape = rlz.shape_like("args")


@public
class ArrayPosition(Value):
    """Return the position of a specific value in an array."""

    arg: Value[dt.Array]
    other: Value

    dtype = dt.int64
    shape = rlz.shape_like("args")


@public
class ArrayRemove(Value):
    """Remove an element from an array."""

    arg: Value[dt.Array]
    other: Value

    dtype = rlz.dtype_like("arg")
    shape = rlz.shape_like("args")


@public
class ArrayDistinct(Value):
    """Return the unique elements of an array."""

    arg: Value[dt.Array]

    dtype = rlz.dtype_like("arg")
    shape = rlz.shape_like("arg")


@public
class ArraySort(Value):
    """Sort the values of an array."""

    arg: Value[dt.Array]

    dtype = rlz.dtype_like("arg")
    shape = rlz.shape_like("arg")


@public
class ArrayUnion(Value):
    """Return the union of two arrays."""

    left: Value[dt.Array]
    right: Value[dt.Array]

    dtype = rlz.dtype_like("args")
    shape = rlz.shape_like("args")


@public
class ArrayIntersect(Value):
    """Return the intersection of two arrays."""

    left: Value[dt.Array]
    right: Value[dt.Array]

    dtype = rlz.dtype_like("args")
    shape = rlz.shape_like("args")


@public
class ArrayZip(Value):
    """Zip two or more arrays into an array of structs."""

    arg: VarTuple[Value[dt.Array]]

    shape = rlz.shape_like("arg")

    @attribute
    def dtype(self):
        return dt.Array(
            dt.Struct(
                {
                    f"f{i:d}": array.dtype.value_type
                    for i, array in enumerate(self.arg, start=1)
                }
            )
        )


@public
class ArrayFlatten(Value):
    """Flatten a nested array one level.

    The input expression must have at least one level of nesting for flattening
    to make sense.
    """

    arg: Value[dt.Array[dt.Array]]
    shape = rlz.shape_like("arg")

    @property
    def dtype(self):
        return self.arg.dtype.value_type


class Range(Value):
    """Base class for range-generating operations."""

    shape = rlz.shape_like("args")

    @attribute
    def dtype(self) -> dt.DataType:
        return dt.Array(dt.highest_precedence((self.start.dtype, self.stop.dtype)))


@public
class IntegerRange(Range):
    """Produce an array of integers from `start` to `stop`, moving by `step`."""

    start: Value[dt.Integer]
    stop: Value[dt.Integer]
    step: Value[dt.Integer]


@public
class TimestampRange(Range):
    """Produce an array of timestamps from `start` to `stop`, moving by `step`."""

    start: Value[dt.Timestamp]
    stop: Value[dt.Timestamp]
    step: Value[dt.Interval]


class ArrayAgg(Value):
    arg: Value[dt.Array]
    shape = rlz.shape_like("args")

    @attribute
    def dtype(self) -> dt.DataType:
        return self.arg.dtype.value_type


@public
class ArrayMin(ArrayAgg):
    """Compute the minimum value of an array."""


@public
class ArrayMax(ArrayAgg):
    """Compute the maximum value of an array."""


# in duckdb summing an array of ints leads to an int, but for other backends
# it might lead to a float??
@public
class ArraySum(ArrayAgg):
    """Compute the sum of an array."""

    arg: Value[dt.Array[dt.Numeric]]


@public
class ArrayMean(ArrayAgg):
    """Compute the average of an array."""

    arg: Value[dt.Array[dt.Numeric]]

    @attribute
    def dtype(self) -> dt.DataType:
        dtype = self.arg.dtype.value_type
        if dtype.is_floating() or dtype.is_integer():
            return dt.float64
        # do nothing for decimal types
        return dtype


@public
class ArrayAny(ArrayAgg):
    """Compute whether any array element is true."""

    arg: Value[dt.Array[dt.Boolean]]


@public
class ArrayAll(ArrayAgg):
    """Compute whether all array elements are true."""

    arg: Value[dt.Array[dt.Boolean]]



import tempfile
from pathlib import Path

import pytest

import xorq.api as xo
import xorq.expr.datatypes as dt
import xorq.vendor.ibis.expr.operations as ops
from xorq.common.utils.graph_utils import gen_children_of, to_node
from xorq.common.utils.lineage_utils import (
    GenericNode,
    TextTree,
    _build_column_tree,
    build_column_trees,
    build_tree,
)
from xorq.vendor.ibis.expr.operations.core import Node


@xo.udf.make_pandas_udf(
    schema=xo.schema({"price": float, "discount": float}),
    return_type=dt.float,
    name="calculate_discount_value",
)
def calculate_discount_value(df):
    return df["price"] * df["discount"]


@pytest.fixture
def sample_expression():
    sales_table = xo.memtable(
        {
            "order_id": [1, 2, 1, 2],
            "price": [100.0, 150.0, 200.0, 250.0],
            "discount": [0.1, 0.2, 0.15, 0.1],
        },
        name="sales",
    )

    sales_with_discount = sales_table.mutate(
        discount_value=calculate_discount_value.on_expr(sales_table)
    )

    expr = sales_with_discount.group_by("order_id").agg(
        total_discount=xo._.discount_value.sum(),
        total_price=xo._.price.sum(),
    )

    return expr


def test__build_column_tree_basic(sample_expression):
    node = to_node(sample_expression)

    lineage_tree = _build_column_tree(node)

    assert isinstance(lineage_tree, GenericNode)
    assert lineage_tree.op is not None
    assert isinstance(lineage_tree.children, tuple)

    assert len(lineage_tree.children) > 0

    for child in lineage_tree.children:
        assert isinstance(child, GenericNode)


def test_build_column_trees_and_display(sample_expression):
    column_trees = build_column_trees(sample_expression)

    assert "total_discount" in column_trees
    assert "total_price" in column_trees

    total_discount_tree = column_trees["total_discount"]
    total_price_tree = column_trees["total_price"]

    assert isinstance(total_discount_tree, GenericNode)
    assert isinstance(total_price_tree, GenericNode)

    display_tree = build_tree(total_discount_tree, dedup=True, max_depth=5)

    assert isinstance(display_tree, TextTree)

    display_tree_no_dedup = build_tree(total_discount_tree, dedup=False)
    assert isinstance(display_tree_no_dedup, TextTree)


def test_complete_lineage_for_total_discount_column(sample_expression):
    from xorq.vendor.ibis.expr.operations.reductions import Sum

    column_trees = build_column_trees(sample_expression)
    total_discount_tree = column_trees["total_discount"]

    def gen_lineage_nodes(node):
        yield node.op
        for child in node.children:
            yield from gen_lineage_nodes(child)

    lineage_nodes = list(gen_lineage_nodes(total_discount_tree))

    # Expected lineage based on actual structure:
    # Sum -> Field(discount_value) -> UDF(calculate_discount_value) -> Field(price/discount) -> InMemoryTable

    node_types = [type(node).__name__ for node in lineage_nodes]

    sum_nodes = [node for node in lineage_nodes if isinstance(node, Sum)]
    assert len(sum_nodes) > 0, (
        f"Should contain Sum operation, found types: {node_types}"
    )

    udf_nodes = [
        node
        for node in lineage_nodes
        if type(node).__name__ == "calculate_discount_value"
    ]
    assert len(udf_nodes) > 0, (
        f"Should contain calculate_discount_value UDF in lineage, found types: {node_types}"
    )

    field_nodes = [node for node in lineage_nodes if isinstance(node, ops.Field)]
    field_names = [node.name for node in field_nodes]

    assert "discount_value" in field_names, (
        f"Should find discount_value field, found: {field_names}"
    )

    source_fields = {"price", "discount"}
    found_source_fields = source_fields.intersection(set(field_names))
    assert len(found_source_fields) > 0, (
        f"Should find at least one source field {source_fields}, found: {field_names}"
    )

    table_nodes = [
        node
        for node in lineage_nodes
        if isinstance(node, (ops.InMemoryTable, ops.UnboundTable, ops.DatabaseTable))
    ]
    assert len(table_nodes) > 0, "Should trace back to original table"


@pytest.fixture
def multi_join_expression():
    """Build a 4-table join expression that creates nested JoinChains.

    Simulates the real-world pattern where sub-expressions are loaded from the
    catalog (YAML round-trip) then joined via ``into_backend``.  The resulting
    expression has 3 nested JoinChains whose shared sub-graphs are traversed
    exponentially by ``_build_column_tree`` (which lacks memoization).
    """
    from xorq.ibis_yaml.compiler import build_expr, load_expr

    batting = xo.memtable(
        {
            "playerID": ["a", "b"],
            "yearID": [2007, 1975],
            "teamID": ["CHA", "MIL"],
            "G": [25, 137],
            "AB": [0, 465],
            "H": [0, 109],
        },
        name="batting",
    )
    people_mem = xo.memtable(
        {
            "playerID": ["a", "b"],
            "nameFirst": ["David", "Hank"],
            "nameLast": ["Aardsma", "Aaron"],
        },
        name="people",
    )
    salaries_mem = xo.memtable(
        {
            "playerID": ["a", "b"],
            "yearID": [2007, 1975],
            "teamID": ["CHA", "MIL"],
            "salary": [387500, 240000],
        },
        name="salaries",
    )
    fielding_mem = xo.memtable(
        {
            "playerID": ["a", "b"],
            "yearID": [2007, 1975],
            "teamID": ["CHA", "MIL"],
            "POS": ["P", "LF"],
        },
        name="fielding",
    )

    # YAML round-trip each table (simulates loading from the catalog)
    with tempfile.TemporaryDirectory() as td:
        p_dir = build_expr(people_mem, builds_dir=Path(td) / "p")
        s_dir = build_expr(salaries_mem, builds_dir=Path(td) / "s")
        f_dir = build_expr(fielding_mem, builds_dir=Path(td) / "f")
        people = load_expr(p_dir)
        salaries = load_expr(s_dir)
        fielding = load_expr(f_dir)

    backend = batting._find_backend()
    people_slim = people[["playerID", "nameFirst", "nameLast"]].into_backend(backend)
    salaries_slim = salaries[["playerID", "yearID", "teamID", "salary"]].into_backend(
        backend
    )
    fielding_slim = fielding[["playerID", "yearID", "teamID", "POS"]].into_backend(
        backend
    )

    with_names = (
        batting.filter(batting.AB > 100)
        .join(people_slim, predicates="playerID", how="left")
        .drop("playerID_right")
    )
    with_salary = with_names.join(
        salaries_slim,
        predicates=["playerID", "yearID", "teamID"],
        how="left",
    ).drop("playerID_right", "yearID_right", "teamID_right")
    return with_salary.join(
        fielding_slim,
        predicates=["playerID", "yearID", "teamID"],
        how="left",
    ).drop("playerID_right", "yearID_right", "teamID_right")


def _count_unique_dag_nodes(expr):
    """Count unique nodes reachable via gen_children_of."""
    visited = set()
    stack = [to_node(expr)]
    while stack:
        n = stack.pop()
        if id(n) not in visited:
            visited.add(id(n))
            stack.extend(gen_children_of(n))
    return len(visited)


def _count_unique_generic_nodes(root: GenericNode) -> int:
    seen, stack = set(), [root]
    while stack:
        n = stack.pop()
        if id(n) not in seen:
            seen.add(id(n))
            stack.extend(n.children)
    return len(seen)


def _build_column_tree_memoized(node: Node, _memo: dict | None = None) -> GenericNode:
    """Gold-standard recursive implementation with id-keyed memoization."""
    if _memo is None:
        _memo = {}
    key = id(node)
    if key in _memo:
        return _memo[key]

    match node:
        case ops.Field(rel=ops.Project(values=values)) as field_node:
            # include the field and recurse into its mapped expression
            mapped = values[field_node.name]
            child = _build_column_tree_memoized(to_node(mapped), _memo)
            result = GenericNode(op=field_node, children=(child,))

        case ops.Field() as field_node:
            children = tuple(
                _build_column_tree_memoized(to_node(child), _memo)
                for child in gen_children_of(field_node)
            )
            result = GenericNode(op=field_node, children=children)

        case ops.Project() as proj:
            result = _build_column_tree_memoized(to_node(proj.parent), _memo)

        case _:
            children = tuple(
                _build_column_tree_memoized(to_node(child), _memo)
                for child in gen_children_of(node)
            )
            result = GenericNode(op=node, children=children)

    _memo[key] = result
    return result


def _assert_matches_gold(expr):
    op = to_node(expr)
    cols = getattr(op, "values", None) or getattr(op, "fields", {})
    gold_memo = {}
    assert len(cols) > 0, (
        f"Expression has no columns to compare; got {type(to_node(expr)).__name__}"
    )
    for name, v in cols.items():
        node = to_node(v)
        assert _build_column_tree(node) == _build_column_tree_memoized(
            node, gold_memo
        ), f"Column '{name}': graph-based result differs from memoized gold standard"
    assert _build_column_tree(op) == _build_column_tree_memoized(op, gold_memo), (
        "graph-based result for the full expression root differs from memoized gold standard"
    )


def test_build_column_tree_matches_memoized_on_sample(sample_expression):
    _assert_matches_gold(sample_expression)


@pytest.mark.benchmark
def test_build_column_tree_matches_memoized_on_multi_join(multi_join_expression):
    _assert_matches_gold(multi_join_expression)


def test_build_column_tree_output_size_bounded(multi_join_expression):
    node = to_node(multi_join_expression)
    result = _build_column_tree(node)
    unique_input = _count_unique_dag_nodes(multi_join_expression)
    unique_output = _count_unique_generic_nodes(result)
    assert unique_output <= unique_input, (
        f"Output DAG has {unique_output} unique GenericNodes for {unique_input} "
        f"unique input nodes — suggests repeated construction"
    )



# https://docs.pytest.org/en/7.1.x/example/parametrize.html#parametrizing-conditional-raising
import shutil
from contextlib import nullcontext as does_not_raise
from pathlib import Path

import pytest
from click.testing import CliRunner

from xorq.catalog.catalog import (
    BuildTgz,
    Catalog,
    CatalogAddition,
)
from xorq.catalog.cli import cli
from xorq.catalog.tar_utils import (
    REQUIRED_TGZ_NAMES,
    extract_build_tgz_context,
    write_tgz,
)
from xorq.catalog.tests.conftest import (
    compare_repo_and_catalog,
    make_build_tgz,
)


@pytest.fixture
def runner():
    yield CliRunner()


# --- init command ---


def test_init_command(runner, tmpdir):
    new_path = str(Path(tmpdir).joinpath("init-catalog"))
    result = runner.invoke(cli, ["--path", new_path, "init"])
    assert result.exit_code == 0, result.output
    assert "Initialized catalog at" in result.output
    assert Path(new_path).exists()


def test_init_command_already_exists(runner, catalog_path):
    assert Path(catalog_path).exists()
    result = runner.invoke(cli, ["--path", catalog_path, "init"])
    assert result.exit_code != 0
    assert "already exists" in result.output
    assert catalog_path in result.output


def test_init_by_name(runner, tmpdir, monkeypatch):
    monkeypatch.setattr(Catalog, "by_name_base_path", Path(tmpdir))
    result = runner.invoke(cli, ["--name", "my-catalog", "init"])
    assert result.exit_code == 0, result.output
    assert "Initialized catalog at" in result.output
    assert Path(tmpdir).joinpath("my-catalog").exists()


def test_init_by_name_already_exists(runner, tmpdir, monkeypatch):
    monkeypatch.setattr(Catalog, "by_name_base_path", Path(tmpdir))
    runner.invoke(cli, ["--name", "my-catalog", "init"])
    result = runner.invoke(cli, ["--name", "my-catalog", "init"])
    assert result.exit_code != 0


def test_init_default(runner, tmpdir, monkeypatch):
    monkeypatch.setattr(Catalog, "by_name_base_path", Path(tmpdir))
    result = runner.invoke(cli, ["init"])
    assert result.exit_code == 0, result.output
    assert "Initialized catalog at" in result.output
    assert Path(tmpdir).joinpath("default").exists()


def test_init_default_already_exists(runner, tmpdir, monkeypatch):
    monkeypatch.setattr(Catalog, "by_name_base_path", Path(tmpdir))
    runner.invoke(cli, ["init"])
    result = runner.invoke(cli, ["init"])
    assert result.exit_code != 0


def test_init_with_root_repo_and_name(runner, root_repo):
    result = runner.invoke(
        cli,
        ["--root-repo", root_repo.working_dir, "--name", "sub-catalog", "init"],
    )
    assert result.exit_code == 0, result.output
    assert "Initialized catalog at" in result.output


# --- add command ---


def test_add_command(runner, catalog_path, data_dict):
    path = str(next(iter(data_dict.values())))
    result = runner.invoke(cli, ["--path", catalog_path, "add", path])
    assert result.exit_code == 0, result.output
    assert "Added" in result.output


def test_add_multiple(runner, catalog_path, data_dict):
    paths = [str(p) for p in data_dict.values()]
    result = runner.invoke(cli, ["--path", catalog_path, "add", *paths])
    assert result.exit_code == 0, result.output
    assert result.output.count("Added") == len(data_dict)


def test_add_with_aliases(runner, catalog_path, data_dict):
    path = str(next(iter(data_dict.values())))
    result = runner.invoke(
        cli,
        [
            "--path",
            catalog_path,
            "add",
            path,
            "--alias",
            "alias-x",
            "--alias",
            "alias-y",
        ],
    )
    assert result.exit_code == 0, result.output
    catalog = Catalog(repo=Catalog.from_kwargs(path=catalog_path, init=False).repo)
    assert {ca.alias for ca in catalog.catalog_aliases} == {"alias-x", "alias-y"}


def test_add_duplicate(runner, catalog_path, data_dict):
    path = str(next(iter(data_dict.values())))
    runner.invoke(cli, ["--path", catalog_path, "add", path])
    result = runner.invoke(cli, ["--path", catalog_path, "add", path])
    assert result.exit_code != 0


def test_add_from_directory(runner, catalog_path, tmpdir):
    tgz = make_build_tgz(tmpdir, "build-dir-test")
    with extract_build_tgz_context(tgz) as build_dir:
        result = runner.invoke(cli, ["--path", catalog_path, "add", str(build_dir)])
    assert result.exit_code == 0, result.output
    assert "Added" in result.output


def test_add_nonexistent_path(runner, catalog_path):
    result = runner.invoke(
        cli, ["--path", catalog_path, "add", "/nonexistent/file.tgz"]
    )
    assert result.exit_code != 0


@pytest.mark.parametrize(
    "sync,expectation",
    (
        (True, does_not_raise()),
        (False, pytest.raises(AssertionError)),
    ),
)
def test_add_sync(sync, expectation, runner, repo_cloned_bare, tmpdir):
    cloned = Catalog.clone_from(
        repo_cloned_bare.working_dir, Path(tmpdir).joinpath("add-sync-test")
    )

    path = make_build_tgz(tmpdir, "to-add")
    result = runner.invoke(
        cli,
        [
            "--path",
            cloned.repo_path,
            "add",
            str(path),
            "--sync" if sync else "--no-sync",
        ],
    )
    assert result.exit_code == 0, result.output
    assert "Added" in result.output

    # check sync condition
    with expectation:
        compare_repo_and_catalog(repo_cloned_bare, cloned)


# --- remove command ---


def test_remove_command(runner, catalog_path, data_dict):
    path = str(next(iter(data_dict.values())))
    runner.invoke(cli, ["--path", catalog_path, "add", path])
    name = Path(path).name.removesuffix("".join(Path(path).suffixes))
    result = runner.invoke(cli, ["--path", catalog_path, "remove", name])
    assert result.exit_code == 0, result.output
    assert "Removed" in result.output


def test_remove_nonexistent(runner, catalog_path):
    result = runner.invoke(cli, ["--path", catalog_path, "remove", "nonexistent"])
    assert result.exit_code != 0


def test_remove_multiple(runner, catalog_path, data_dict):
    paths = [str(p) for p in data_dict.values()]
    runner.invoke(cli, ["--path", catalog_path, "add", *paths])
    names = [
        Path(p).name.removesuffix("".join(Path(p).suffixes)) for p in data_dict.values()
    ]
    result = runner.invoke(cli, ["--path", catalog_path, "remove", *names])
    assert result.exit_code == 0, result.output
    assert result.output.count("Removed") == len(data_dict)


@pytest.mark.parametrize(
    "sync,expectation",
    (
        (True, does_not_raise()),
        (False, pytest.raises(AssertionError)),
    ),
)
def test_remove_sync(sync, expectation, runner, repo_cloned_bare, tmpdir, data_dict):
    cloned = Catalog.clone_from(
        repo_cloned_bare.working_dir, Path(tmpdir).joinpath("remove-sync-test")
    )

    path = str(next(iter(data_dict.values())))
    result = runner.invoke(
        cli,
        [
            "--path",
            cloned.repo_path,
            "remove",
            Path(path).stem,
            "--sync" if sync else "--no-sync",
        ],
    )
    assert result.exit_code == 0, result.output
    assert "Removed" in result.output

    # check that sync condition
    with expectation:
        compare_repo_and_catalog(repo_cloned_bare, cloned)


# --- add-alias command ---


def test_add_alias_command(runner, catalog_path, data_dict):
    path = str(next(iter(data_dict.values())))
    runner.invoke(cli, ["--path", catalog_path, "add", path])
    name = Path(path).name.removesuffix("".join(Path(path).suffixes))
    result = runner.invoke(cli, ["--path", catalog_path, "add-alias", name, "my-alias"])
    assert result.exit_code == 0, result.output
    assert "my-alias" in result.output


def test_add_with_aliases_commit_message(runner, catalog_path, data_dict):
    path = str(next(iter(data_dict.values())))
    result = runner.invoke(
        cli,
        ["--path", catalog_path, "add", path, "--alias", "v1", "--alias", "latest"],
    )
    assert result.exit_code == 0, result.output
    catalog = Catalog.from_kwargs(path=catalog_path, init=False)
    commit_message = catalog.repo.head.commit.message.strip()
    assert "v1" in commit_message
    assert "latest" in commit_message


def test_add_alias_unknown_entry(runner, catalog_path):
    result = runner.invoke(
        cli, ["--path", catalog_path, "add-alias", "nonexistent", "my-alias"]
    )
    assert result.exit_code != 0


def test_add_alias_overwrite(runner, catalog_path, data_dict):
    paths = [str(p) for p in data_dict.values()]
    runner.invoke(cli, ["--path", catalog_path, "add", *paths])
    names = [
        Path(p).name.removesuffix("".join(Path(p).suffixes)) for p in data_dict.values()
    ]
    runner.invoke(cli, ["--path", catalog_path, "add-alias", names[0], "shared"])
    result = runner.invoke(
        cli, ["--path", catalog_path, "add-alias", names[1], "shared"]
    )
    assert result.exit_code == 0, result.output
    assert "shared" in result.output


# --- remove-alias command ---


def test_remove_alias_command(runner, catalog_path, data_dict):
    path = str(next(iter(data_dict.values())))
    runner.invoke(cli, ["--path", catalog_path, "add", path])
    name = Path(path).name.removesuffix("".join(Path(path).suffixes))
    runner.invoke(cli, ["--path", catalog_path, "add-alias", name, "to-remove"])
    result = runner.invoke(cli, ["--path", catalog_path, "remove-alias", "to-remove"])
    assert result.exit_code == 0, result.output
    assert "to-remove" in result.output


def test_remove_alias_multiple(runner, catalog_path, data_dict):
    path = str(next(iter(data_dict.values())))
    runner.invoke(cli, ["--path", catalog_path, "add", path])
    name = Path(path).name.removesuffix("".join(Path(path).suffixes))
    runner.invoke(cli, ["--path", catalog_path, "add-alias", name, "alias-a"])
    runner.invoke(cli, ["--path", catalog_path, "add-alias", name, "alias-b"])
    result = runner.invoke(
        cli, ["--path", catalog_path, "remove-alias", "alias-a", "alias-b"]
    )
    assert result.exit_code == 0, result.output
    assert result.output.count("Removed alias") == 2


def test_remove_alias_nonexistent(runner, catalog_path):
    result = runner.invoke(
        cli, ["--path", catalog_path, "remove-alias", "no-such-alias"]
    )
    assert result.exit_code != 0


@pytest.mark.parametrize(
    "sync,expectation",
    (
        (True, does_not_raise()),
        (False, pytest.raises(AssertionError)),
    ),
)
def test_add_alias_sync(sync, expectation, runner, repo_cloned_bare, tmpdir):
    cloned = Catalog.clone_from(
        repo_cloned_bare.working_dir, Path(tmpdir).joinpath("add-alias-sync-test")
    )
    path = make_build_tgz(tmpdir, "to-alias")
    runner.invoke(cli, ["--path", str(cloned.repo_path), "add", str(path), "--sync"])
    name = path.stem
    result = runner.invoke(
        cli,
        [
            "--path",
            str(cloned.repo_path),
            "add-alias",
            name,
            "my-alias",
            "--sync" if sync else "--no-sync",
        ],
    )
    assert result.exit_code == 0, result.output
    assert "my-alias" in result.output

    with expectation:
        compare_repo_and_catalog(repo_cloned_bare, cloned)


@pytest.mark.parametrize(
    "sync,expectation",
    (
        (True, does_not_raise()),
        (False, pytest.raises(AssertionError)),
    ),
)
def test_remove_alias_sync(sync, expectation, runner, repo_cloned_bare, tmpdir):
    cloned = Catalog.clone_from(
        repo_cloned_bare.working_dir, Path(tmpdir).joinpath("remove-alias-sync-test")
    )
    path = make_build_tgz(tmpdir, "to-alias")
    runner.invoke(cli, ["--path", str(cloned.repo_path), "add", str(path), "--sync"])
    name = path.stem
    runner.invoke(
        cli,
        ["--path", str(cloned.repo_path), "add-alias", name, "my-alias", "--sync"],
    )
    result = runner.invoke(
        cli,
        [
            "--path",
            str(cloned.repo_path),
            "remove-alias",
            "my-alias",
            "--sync" if sync else "--no-sync",
        ],
    )
    assert result.exit_code == 0, result.output
    assert "my-alias" in result.output

    with expectation:
        compare_repo_and_catalog(repo_cloned_bare, cloned)


def test_remove_entry_cascades_aliases(runner, catalog_path, data_dict):
    path = str(next(iter(data_dict.values())))
    runner.invoke(cli, ["--path", catalog_path, "add", path])
    name = Path(path).name.removesuffix("".join(Path(path).suffixes))
    runner.invoke(cli, ["--path", catalog_path, "add-alias", name, "alias-p"])
    runner.invoke(cli, ["--path", catalog_path, "add-alias", name, "alias-q"])

    result = runner.invoke(cli, ["--path", catalog_path, "remove", name])
    assert result.exit_code == 0, result.output

    catalog = Catalog.from_kwargs(path=catalog_path, init=False)
    assert catalog.list_aliases() == []
    catalog.assert_consistency()


# --- info command ---


def test_info_empty(runner, catalog_path):
    result = runner.invoke(cli, ["--path", catalog_path, "info"])
    assert result.exit_code == 0, result.output
    assert "path:" in result.output
    assert "commit:" in result.output
    assert "entries: 0" in result.output
    assert "aliases: 0" in result.output
    assert "(none)" in result.output


def test_info_populated(runner, catalog_path, data_dict):
    paths = [str(p) for p in data_dict.values()]
    runner.invoke(cli, ["--path", catalog_path, "add", *paths])
    result = runner.invoke(cli, ["--path", catalog_path, "info"])
    assert result.exit_code == 0, result.output
    assert f"entries: {len(data_dict)}" in result.output
    assert "aliases: 0" in result.output


def test_info_alias_count(runner, catalog_path, data_dict):
    path = str(next(iter(data_dict.values())))
    runner.invoke(cli, ["--path", catalog_path, "add", path])
    name = Path(path).name.removesuffix("".join(Path(path).suffixes))
    runner.invoke(cli, ["--path", catalog_path, "add-alias", name, "alias-p"])
    runner.invoke(cli, ["--path", catalog_path, "add-alias", name, "alias-q"])
    result = runner.invoke(cli, ["--path", catalog_path, "info"])
    assert result.exit_code == 0, result.output
    assert "aliases: 2" in result.output


def test_info_shows_remotes(runner, repo_cloned_bare, tmpdir):
    cloned = Catalog.clone_from(
        repo_cloned_bare.working_dir, Path(tmpdir).joinpath("info-remote-test")
    )
    result = runner.invoke(cli, ["--path", str(cloned.repo_path), "info"])
    assert result.exit_code == 0, result.output
    assert "origin" in result.output


# --- list command ---


def test_list_empty(runner, catalog_path):
    result = runner.invoke(cli, ["--path", catalog_path, "list"])
    assert result.exit_code == 0, result.output
    assert "No entries." in result.output


def test_list_populated(runner, catalog_path, data_dict):
    paths = [str(p) for p in data_dict.values()]
    runner.invoke(cli, ["--path", catalog_path, "add", *paths])
    result = runner.invoke(cli, ["--path", catalog_path, "list"])
    assert result.exit_code == 0, result.output
    assert "No entries." not in result.output
    for p in data_dict.values():
        name = Path(p).name.removesuffix("".join(Path(p).suffixes))
        assert name in result.output


# --- list-aliases command ---


def test_list_aliases_invalid_path(runner):
    result = runner.invoke(cli, ["--path", "/nonexistent/path", "list-aliases"])
    assert result.exit_code != 0
    assert "init" in result.output


def test_list_aliases_empty(runner, catalog_path):
    result = runner.invoke(cli, ["--path", catalog_path, "list-aliases"])
    assert result.exit_code == 0, result.output
    assert "No aliases." in result.output


def test_list_aliases_populated(runner, catalog_path, data_dict):
    path = str(next(iter(data_dict.values())))
    runner.invoke(cli, ["--path", catalog_path, "add", path])
    name = Path(path).name.removesuffix("".join(Path(path).suffixes))
    runner.invoke(cli, ["--path", catalog_path, "add-alias", name, "alias-a"])
    runner.invoke(cli, ["--path", catalog_path, "add-alias", name, "alias-b"])
    result = runner.invoke(cli, ["--path", catalog_path, "list-aliases"])
    assert result.exit_code == 0, result.output
    assert "alias-a" in result.output
    assert "alias-b" in result.output


# --- get command ---


def test_get_command(runner, catalog_path, data_dict, tmpdir):
    path = str(next(iter(data_dict.values())))
    runner.invoke(cli, ["--path", catalog_path, "add", path])
    name = Path(path).name.removesuffix("".join(Path(path).suffixes))
    output_dir = str(Path(tmpdir).joinpath("export"))
    Path(output_dir).mkdir()
    result = runner.invoke(cli, ["--path", catalog_path, "get", name, "-o", output_dir])
    assert result.exit_code == 0, result.output
    assert "Exported to" in result.output


def test_get_default_output_dir(runner, catalog_path, data_dict):
    path = str(next(iter(data_dict.values())))
    runner.invoke(cli, ["--path", catalog_path, "add", path])
    name = Path(path).name.removesuffix("".join(Path(path).suffixes))
    with runner.isolated_filesystem():
        result = runner.invoke(cli, ["--path", catalog_path, "get", name])
        assert result.exit_code == 0, result.output
        assert "Exported to" in result.output


def test_get_nonexistent_entry(runner, catalog_path, tmpdir):
    output_dir = str(Path(tmpdir).joinpath("export"))
    Path(output_dir).mkdir()
    result = runner.invoke(
        cli, ["--path", catalog_path, "get", "nonexistent", "-o", output_dir]
    )
    assert result.exit_code != 0


# --- check command ---


def test_check_command(runner, catalog_path):
    result = runner.invoke(cli, ["--path", catalog_path, "check"])
    assert result.exit_code == 0, result.output
    assert "OK" in result.output


def test_check_catches_inconsistency(runner, catalog_path, tmpdir):
    catalog = Catalog.from_kwargs(path=catalog_path, init=False)
    tgz_path = write_tgz(
        Path(tmpdir).joinpath("build.tgz"),
        dict.fromkeys(REQUIRED_TGZ_NAMES, b""),
    )
    catalog_addition = CatalogAddition(BuildTgz(tgz_path), catalog)
    catalog_addition.ensure_dirs()
    entry_path = catalog_addition.catalog_entry.catalog_path
    with catalog.commit_context("bad commit"):
        shutil.copy(tgz_path, entry_path)
        catalog.repo.index.add((entry_path,))

    result = runner.invoke(cli, ["--path", catalog_path, "check"])
    assert result.exit_code != 0


def test_check_populated(runner, catalog_path, data_dict):
    paths = [str(p) for p in data_dict.values()]
    runner.invoke(cli, ["--path", catalog_path, "add", *paths])
    result = runner.invoke(cli, ["--path", catalog_path, "check"])
    assert result.exit_code == 0, result.output
    assert "OK" in result.output


# --- clone command ---


def test_clone_command(runner, catalog_populated, tmpdir):
    dest = str(Path(tmpdir).joinpath("cloned"))
    result = runner.invoke(
        cli, ["clone", str(catalog_populated.repo_path), "--path", dest]
    )
    assert result.exit_code == 0, result.output
    assert "Cloned to" in result.output
    assert Path(dest).exists()


def test_clone_name_and_path_mutually_exclusive(runner, catalog_populated):
    result = runner.invoke(
        cli,
        [
            "clone",
            str(catalog_populated.repo_path),
            "--name",
            "foo",
            "--path",
            "/tmp/bar",
        ],
    )
    assert result.exit_code != 0


def test_clone_derives_name_from_url(runner, catalog_populated, tmpdir, monkeypatch):
    clone_base = Path(tmpdir).joinpath("clone-base")
    clone_base.mkdir()
    monkeypatch.setattr(Catalog, "by_name_base_path", clone_base)
    result = runner.invoke(cli, ["clone", str(catalog_populated.repo_path)])
    assert result.exit_code == 0, result.output
    assert "Cloned to" in result.output


# --- push / pull / sync ---


def test_push_with_remote(runner, repo_cloned_bare, tmpdir):
    cloned = Catalog.clone_from(
        repo_cloned_bare.working_dir, Path(tmpdir).joinpath("push-test")
    )
    compare_repo_and_catalog(repo_cloned_bare, cloned)
    before = repo_cloned_bare.head.commit.hexsha

    result = runner.invoke(cli, ["--path", str(cloned.repo_path), "push"])
    assert result.exit_code == 0, result.output
    assert "Pushed." in result.output
    middle = repo_cloned_bare.head.commit.hexsha
    assert before == middle

    # check commit
    path = make_build_tgz(tmpdir, "to-add")
    result = runner.invoke(
        cli, ["--path", cloned.repo_path, "add", str(path), "--no-sync"]
    )
    assert result.exit_code == 0, result.output
    assert "Added" in result.output
    middle = repo_cloned_bare.head.commit.hexsha
    assert before == middle
    with pytest.raises(AssertionError):
        compare_repo_and_catalog(repo_cloned_bare, cloned)

    result = runner.invoke(cli, ["--path", str(cloned.repo_path), "push"])
    assert result.exit_code == 0, result.output
    assert "Pushed." in result.output
    after = repo_cloned_bare.head.commit.hexsha
    assert before != after
    compare_repo_and_catalog(repo_cloned_bare, cloned)


def test_pull_with_remote(runner, repo_cloned_bare, tmpdir):
    cloned = Catalog.clone_from(
        repo_cloned_bare.working_dir, Path(tmpdir).joinpath("pull-test")
    )
    pusher = Catalog.clone_from(
        repo_cloned_bare.working_dir, Path(tmpdir).joinpath("pusher")
    )
    before = cloned.repo.head.commit.hexsha

    result = runner.invoke(cli, ["--path", str(cloned.repo_path), "pull"])
    assert result.exit_code == 0, result.output
    assert "Pulled." in result.output
    #
    middle = cloned.repo.head.commit.hexsha
    assert before == middle
    assert middle == pusher.repo.head.commit.hexsha

    path = make_build_tgz(tmpdir, "to-add")
    pusher.add(path, sync=True)
    middle = cloned.repo.head.commit.hexsha
    assert before == middle
    assert middle != pusher.repo.head.commit.hexsha

    result = runner.invoke(cli, ["--path", str(cloned.repo_path), "pull"])
    assert result.exit_code == 0, result.output
    assert "Pulled." in result.output
    #
    after = cloned.repo.head.commit.hexsha
    assert before != after
    assert after == pusher.repo.head.commit.hexsha


def test_sync_with_remote(runner, repo_cloned_bare, tmpdir):
    cloned = Catalog.clone_from(
        repo_cloned_bare.working_dir, Path(tmpdir).joinpath("sync-test")
    )
    result = runner.invoke(cli, ["--path", str(cloned.repo_path), "sync"])
    assert result.exit_code == 0, result.output
    assert "Synced." in result.output


# --- CLI option validation ---


def test_help(runner):
    result = runner.invoke(cli, ["--help"])
    assert result.exit_code == 0
    assert "Manage xorq build-artifact catalogs." in result.output


def test_subcommand_help(runner):
    for cmd in (
        "init",
        "add",
        "add-alias",
        "remove",
        "remove-alias",
        "info",
        "list",
        "list-aliases",
        "get",
        "push",
        "pull",
        "sync",
        "clone",
        "check",
    ):
        result = runner.invoke(cli, [cmd, "--help"])
        assert result.exit_code == 0, f"{cmd} --help failed"


def test_name_and_path_mutually_exclusive(runner, catalog_path, tmpdir, monkeypatch):
    monkeypatch.setattr(Catalog, "by_name_base_path", Path(tmpdir))
    runner.invoke(cli, ["--name", "my-catalog", "init"])
    result = runner.invoke(
        cli, ["--name", "my-catalog", "--path", catalog_path, "list"]
    )
    assert result.exit_code != 0


def test_list_invalid_path(runner):
    result = runner.invoke(cli, ["--path", "/nonexistent/path", "list"])
    assert result.exit_code != 0
    assert "init" in result.output


def test_missing_catalog_hint_includes_path(runner, tmpdir):
    missing = str(Path(tmpdir).joinpath("no-such-catalog"))
    result = runner.invoke(cli, ["--path", missing, "list"])
    assert result.exit_code != 0
    assert f"--path {missing}" in result.output
    assert "init" in result.output


def test_missing_catalog_hint_includes_name(runner, tmpdir, monkeypatch):
    monkeypatch.setattr(Catalog, "by_name_base_path", Path(tmpdir))
    result = runner.invoke(cli, ["--name", "my-catalog", "list"])
    assert result.exit_code != 0
    assert "--name my-catalog" in result.output
    assert "init" in result.output


# --- TUI validation ---


def test_tui_missing_catalog_shows_hint(runner, tmpdir):
    missing = str(Path(tmpdir).joinpath("no-such-catalog"))
    result = runner.invoke(cli, ["--path", missing, "tui"])
    assert result.exit_code != 0
    assert "init" in result.output


def test_tui_missing_catalog_with_name(runner, tmpdir):
    result = runner.invoke(cli, ["--name", "no-such-catalog-xyz", "tui"])
    assert result.exit_code != 0
    assert "init" in result.output



import pandas as pd

import xorq.api as xo
from xorq.expr.datatypes import Int64, LargeString, String
from xorq.tests.util import assert_frame_equal, assert_series_equal
from xorq.vendor import ibis


def test_can_create_table(utf8_data):
    con = xo.connect()
    schema = ibis.schema([("name", LargeString), ("age", Int64)])
    t = con.create_table("names", utf8_data, schema=schema)
    assert t.schema() == schema
    assert "names" in con.list_tables()

    actual = xo.execute(t)  # can be executed
    assert_frame_equal(utf8_data, actual)


def test_can_read_table(utf8_data):
    con = xo.connect()

    data = """
    create table utf8_data(str string, val bigint) as values
      ('A', 1),
      ('B', 2),
      ('A', 2),
      ('A', 4),
      ('C', 1),
      ('A', 1);
    """

    con.con.sql(data)

    large_data = """
    create table largeutf8_data as
    select arrow_cast(str, 'LargeUtf8') as str, val
    from utf8_data;
    """

    con.con.sql(large_data)

    t = con.table("largeutf8_data")
    schema = ibis.schema([("str", LargeString), ("val", Int64)])
    assert t.schema() == schema

    expr = t.select(t.str.name("name"), t.val.name("age"))
    actual = xo.execute(expr)

    assert_frame_equal(actual, utf8_data)


def test_can_read_write_parquet(utf8_data, tmp_path):
    con = xo.connect()
    schema = ibis.schema([("name", LargeString), ("age", Int64)])
    t = con.create_table("names", utf8_data, schema=schema)

    names_parquet_path = tmp_path / "names.parquet"
    xo.to_parquet(t, names_parquet_path)

    # need to specify the schema when reading parquet because DataFusion transforms into StringView
    t = xo.read_parquet(names_parquet_path, "names_parquet", schema=schema.to_pyarrow())
    assert t.schema() == schema

    actual = xo.execute(t)

    assert_frame_equal(utf8_data, actual)


def test_can_execute_test_ops(utf8_data):
    con = xo.connect()
    schema = ibis.schema([("name", LargeString), ("age", Int64)])
    t = con.create_table("names", utf8_data, schema=schema)
    assert t.schema() == schema

    expr = t.name.startswith("A").name("prefix")
    actual = con.execute(expr)
    expected = utf8_data["name"].str.startswith("A")

    assert_series_equal(actual, expected, check_names=False)


def test_cast(utf8_data):
    con = xo.connect()
    t = con.register(utf8_data, "t")
    expr = t.mutate(a=t.name.cast(LargeString))
    actual = xo.execute(expr)

    assert expr.schema() == ibis.schema(
        [("name", String), ("age", Int64), ("a", LargeString)]
    )
    assert isinstance(actual, pd.DataFrame)



"""Initialize Ibis module."""

from __future__ import annotations


__version__ = "9.5.0"

import warnings
from typing import Any

from xorq.common.exceptions import XorqError
from xorq.vendor.ibis import examples, util
from xorq.vendor.ibis.backends import BaseBackend
from xorq.vendor.ibis.config import options
from xorq.vendor.ibis.expr import api
from xorq.vendor.ibis.expr import types as ir
from xorq.vendor.ibis.expr.api import *  # noqa: F403
from xorq.vendor.ibis.expr.operations import udf


__all__ = [  # noqa: PLE0604
    "api",
    "examples",
    "ir",
    "udf",
    "util",
    "BaseBackend",
    "XorqError",
    "options",
    *api.__all__,
]

_KNOWN_BACKENDS = ["heavyai"]


def __dir__() -> list[str]:
    """Adds tab completion for ibis backends to the top-level module."""

    out = set(__all__)
    out.update(ep.name for ep in util.backend_entry_points())
    return sorted(out)


def load_backend(name: str) -> BaseBackend:
    """Load backends in a lazy way with `ibis.<backend-name>`.

    This also registers the backend options.

    Examples
    --------
    >>> import ibis
    >>> con = ibis.sqlite.connect(...)

    When accessing the `sqlite` attribute of the `ibis` module, this function
    is called, and a backend with the `sqlite` name is tried to load from
    the `ibis.backends` entrypoints. If successful, the `ibis.sqlite`
    attribute is "cached", so this function is only called the first time.

    """

    entry_points = {ep for ep in util.backend_entry_points() if ep.name == name}

    if not entry_points:
        msg = f"module 'ibis' has no attribute '{name}'. "
        if name in _KNOWN_BACKENDS:
            msg += f"""If you are trying to access the '{name}' backend,
                    try installing it first with `pip install 'ibis-framework[{name}]'`"""
        raise AttributeError(msg)

    if len(entry_points) > 1:
        raise RuntimeError(
            f"{len(entry_points)} packages found for backend '{name}': "
            f"{entry_points}\n"
            "There should be only one, please uninstall the unused packages "
            "and just leave the one that needs to be used."
        )

    import types

    from xorq.vendor import ibis

    (entry_point,) = entry_points
    try:
        module = entry_point.load()
    except ImportError as exc:
        raise ImportError(
            f"Failed to import the {name} backend due to missing dependencies.\n\n"
            f"You can pip or conda install the {name} backend as follows:\n\n"
            f'  python -m pip install -U "ibis-framework[{name}]"  # pip install\n'
            f"  conda install -c conda-forge ibis-{name}           # or conda install"
        ) from exc
    backend = module.Backend()
    # The first time a backend is loaded, we register its options, and we set
    # it as an attribute of `ibis`, so `__getattr__` is not called again for it
    backend.register_options()

    # We don't want to expose all the methods on an unconnected backend to the user.
    # In lieu of a full redesign, we create a proxy module and add only the methods
    # that are valid to call without a connect call. These are:
    #
    # - connect
    # - compile
    # - has_operation
    # - _from_url
    #
    # We also copy over the docstring from `do_connect` to the proxy `connect`
    # method, since that's where all the backend-specific kwargs are currently
    # documented. This is all admittedly gross, but it works and doesn't
    # require a backend redesign yet.

    def connect(*args, **kwargs):
        return backend.connect(*args, **kwargs)

    connect.__doc__ = backend.do_connect.__doc__
    connect.__wrapped__ = backend.do_connect
    connect.__module__ = f"ibis.{name}"

    proxy = types.ModuleType(f"ibis.{name}")
    setattr(ibis, name, proxy)
    proxy.connect = connect
    proxy.compile = backend.compile
    proxy.has_operation = backend.has_operation
    proxy.name = name
    proxy._from_url = backend._from_url

    # Add any additional methods that should be exposed at the top level
    for attr in getattr(backend, "_top_level_methods", ()):
        setattr(proxy, attr, getattr(backend, attr))

    return proxy


def __getattr__(name: str) -> Any:
    if name == "NA":
        warnings.warn(
            "The 'ibis.NA' constant is deprecated as of v9.1 and will be removed in a future "
            "version. Use 'ibis.null()' instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        return null()  # noqa: F405
    else:
        return load_backend(name)



import pytest

import xorq.api as xo
from xorq.caching import ParquetCache


def test_put_get_drop(tmp_path, parquet_dir):
    astronauts_path = parquet_dir.joinpath("astronauts.parquet")

    con = xo.datafusion.connect()
    t = con.read_parquet(astronauts_path, table_name="astronauts")

    cache = ParquetCache.from_kwargs(relative_path=tmp_path, source=con)
    put_node = cache.put(t, t.op())
    assert put_node is not None

    get_node = cache.get(t)
    assert get_node is not None

    cache.drop(t)
    with pytest.raises(KeyError):
        cache.get(t)


def test_default_connection(tmp_path, parquet_dir):
    batting_path = parquet_dir.joinpath("astronauts.parquet")

    con = xo.connect()
    t = con.read_parquet(batting_path, table_name="astronauts")

    # if we do cross source caching, then we get a random name and cache.calc_key result isn't stable
    cache = ParquetCache.from_kwargs(relative_path=tmp_path)
    cache.put(t, t.op())

    get_node = cache.get(t)
    assert get_node is not None
    assert get_node.source.name == con.name
    assert get_node.to_expr().execute is not None



import random

import pytest
from pytest import param

import xorq.api as xo
from xorq.api import _


@pytest.fixture
def union_subsets(alltypes):
    randomizer = random.Random(42)
    cols_a, cols_b, cols_c = (
        randomizer.sample(alltypes.columns, k=len(alltypes.columns)) for _ in range(3)
    )
    assert cols_a != cols_b != cols_c

    a = alltypes.filter((_.id >= 5200) & (_.id <= 5210))[cols_a]
    b = alltypes.filter((_.id >= 5205) & (_.id <= 5215))[cols_b]
    c = alltypes.filter((_.id >= 5213) & (_.id <= 5220))[cols_c]

    return a, b, c


@pytest.mark.parametrize("distinct", [False, True], ids=["all", "distinct"])
def test_union(compiler, union_subsets, distinct):
    a, b, c = union_subsets

    expr = xo.union(a, b, distinct=distinct).order_by("id")
    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)

    assert roundtrip_expr.equals(expr)


def test_union_mixed_distinct(compiler, union_subsets):
    a, b, c = union_subsets

    expr = a.union(b, distinct=True).union(c, distinct=False).order_by("id")
    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)

    assert roundtrip_expr.equals(expr)


@pytest.mark.parametrize(
    "distinct",
    [
        param(False, id="all"),
        param(True, id="distinct"),
    ],
)
def test_intersect(compiler, alltypes, distinct):
    a = alltypes.filter((_.id >= 5200) & (_.id <= 5210))
    b = alltypes.filter((_.id >= 5205) & (_.id <= 5215))
    c = alltypes.filter((_.id >= 5195) & (_.id <= 5208))

    expr = xo.intersect(a, b, c, distinct=distinct).order_by("id")
    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)

    assert roundtrip_expr.equals(expr)


@pytest.mark.parametrize(
    "distinct",
    [
        param(False, id="all"),
        param(True, id="distinct"),
    ],
)
def test_difference(compiler, alltypes, distinct):
    a = alltypes.filter((_.id >= 5200) & (_.id <= 5210))
    b = alltypes.filter((_.id >= 5205) & (_.id <= 5215))
    c = alltypes.filter((_.id >= 5195) & (_.id <= 5202))

    expr = xo.difference(a, b, c, distinct=distinct).order_by("id")
    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)

    assert roundtrip_expr.equals(expr)



"""Operations for working with structs."""

from __future__ import annotations

from public import public

import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.rules as rlz
from xorq.vendor.ibis.common.annotations import ValidationError, attribute
from xorq.vendor.ibis.common.typing import VarTuple  # noqa: TCH001
from xorq.vendor.ibis.expr.operations.core import Value


@public
class StructField(Value):
    """Extract a field from a struct value."""

    arg: Value[dt.Struct]
    field: str

    shape = rlz.shape_like("arg")

    @attribute
    def dtype(self) -> dt.DataType:
        struct_dtype = self.arg.dtype
        value_dtype = struct_dtype[self.field]
        return value_dtype

    @property
    def name(self) -> str:
        return self.field


@public
class StructColumn(Value):
    """Construct a struct column from literals or expressions."""

    names: VarTuple[str]
    values: VarTuple[Value]

    shape = rlz.shape_like("values")

    def __init__(self, names, values):
        if len(names) != len(values):
            raise ValidationError(
                f"Length of names ({len(names)}) does not match length of "
                f"values ({len(values)})"
            )
        super().__init__(names=names, values=values)

    @property
    def name(self) -> str:
        pairs = ", ".join(
            f"{name!r}: {op.name}" for name, op in zip(self.names, self.values)
        )
        return f"{self.__class__.__name__}({{{pairs}}})"

    @attribute
    def dtype(self) -> dt.DataType:
        dtypes = (value.dtype for value in self.values)
        return dt.Struct.from_tuples(zip(self.names, dtypes))



from __future__ import annotations

import os
import sqlite3
from pathlib import Path

import pytest
from pytest import param

import xorq.api as xo
import xorq.vendor.ibis.expr.operations as ops
from xorq.vendor import ibis


def test_attach_file(tmp_path):
    dbpath = str(tmp_path / "attached.db")
    path_client = xo.sqlite.connect(dbpath)
    path_client.create_table("test", schema=ibis.schema({"a": "int"}))

    client = xo.sqlite.connect()

    assert not client.list_tables()

    client.attach("baz", Path(dbpath))
    client.attach("bar", dbpath)

    foo_tables = client.list_tables(database="baz")
    bar_tables = client.list_tables(database="bar")

    assert foo_tables == ["test"]
    assert foo_tables == bar_tables


@pytest.mark.parametrize(
    "url, ext",
    [
        param(lambda p: p, "sqlite", id="no-scheme-sqlite-ext"),
        param(lambda p: p, "db", id="no-scheme-db-ext"),
        param(lambda p: p, "db", id="absolute-path"),
        param(
            lambda p: f"{os.path.relpath(p)}",
            "db",
            id="relative-path",
        ),
        param(lambda _: "", "db", id="in-memory-empty"),
        param(lambda _: ":memory:", "db", id="in-memory-explicit"),
    ],
)
def test_connect(url, ext, tmp_path):
    path = url(os.path.abspath(tmp_path / f"test.{ext}"))
    with sqlite3.connect(path):
        pass
    con = xo.sqlite.connect(path)
    one = xo.literal(1)
    assert con.execute(one) == 1


@pytest.mark.parametrize(
    "op",
    (
        ops.Project,  # Core operations handled in non-standard ways
        ops.Filter,  # Core operations handled in non-standard ways
        ops.Sort,  # Core operations handled in non-standard ways
        ops.Aggregate,  # Core operations handled in non-standard ways
        ops.Capitalize,  # Handled by base class rewrite
        ops.Sample,  # Handled by compiler-specific rewrite
        ops.Cast,  # Handled by visit_* method
    ),
)
def test_has_operation(sqlite_con, op):
    assert sqlite_con.has_operation(op)


def test_list_temp_tables_by_default(sqlite_con):
    name = ibis.util.gen_name("sqlite_temp_table")
    sqlite_con.create_table(name, schema={"a": "int"}, temp=True)
    assert name in sqlite_con.list_tables(database="temp")
    assert name in sqlite_con.list_tables()



import itertools
import pathlib
import re
import types
import urllib.error
import urllib.request

import dask
import sqlglot as sg

import xorq.expr.datatypes as dat
import xorq.expr.relations as rel
import xorq.vendor.ibis.expr.operations.relations as ir
from xorq.common.utils.dask_normalize.dask_normalize_utils import (
    normalize_seq_with_caller,
)
from xorq.expr import api
from xorq.vendor import ibis
from xorq.vendor.ibis.expr.operations.udf import (
    AggUDF,
    InputType,
    ScalarUDF,
)


def expr_is_bound(expr):
    backends, _ = expr._find_backends()
    return bool(backends)


def unbound_expr_to_default_sql(expr, compiler=None):
    if expr_is_bound(expr):
        raise ValueError("expr must be unbound, but is already bound to a backend")
    default_sql = api.to_sql(expr, compiler=compiler)
    return str(default_sql)


def normalize_inmemorytable(dt):
    if not isinstance(dt, ir.InMemoryTable):
        raise ValueError(f"expected InMemoryTable, got {type(dt)}")
    return normalize_seq_with_caller(
        dt.schema.to_pandas(),
        # in memory: so we can assume it's reasonable to hash the data
        tuple(
            dask.base.tokenize(el.serialize().to_pybytes())
            for el in dt.to_expr().to_pyarrow_batches()
        ),
        caller="normalize_inmemorytable",
    )


def normalize_memory_databasetable(dt):
    if dt.source.name not in ("pandas", "xorq", "datafusion", "duckdb", "sqlite"):
        raise ValueError(f"expected in-memory backend, got {dt.source.name!r}")
    return normalize_seq_with_caller(
        # we are normalizing the data, we don't care about the connection
        # dt.source,
        dt.schema.to_pandas(),
        # in memory: so we can assume it's reasonable to hash the data
        tuple(
            dask.base.tokenize(el.serialize().to_pybytes())
            for el in dt.to_expr().to_pyarrow_batches()
        ),
        caller="normalize_memory_databasetable",
    )


def normalize_pandas_databasetable(dt):
    if dt.source.name != "pandas":
        raise ValueError(f"expected pandas backend, got {dt.source.name!r}")
    return normalize_memory_databasetable(dt)


def normalize_datafusion_databasetable(dt):
    if dt.source.name not in ("datafusion", "xorq"):
        raise ValueError(f"expected datafusion/xorq backend, got {dt.source.name!r}")
    table = dt.source.con.table(dt.name)
    ep_str = str(table.execution_plan())
    if ep_str.startswith(("ParquetExec:", "CsvExec:")) or re.match(
        r"DataSourceExec:.+file_type=(csv|parquet)", ep_str
    ):
        return normalize_seq_with_caller(
            dt.schema.to_pandas(),
            # ep_str denotes the parquet files to be read
            # FIXME: md5sum on detected .parquet files?
            ep_str,
        )
    elif ep_str.startswith(("MemoryExec:", "DataSourceExec:")):
        return normalize_memory_databasetable(dt)
    elif "PyRecordBatchProviderExec" in ep_str:
        return normalize_seq_with_caller(
            dt.schema.to_pandas(),
            dt.name,
        )
    elif ep_str.startswith("EmptyExec"):
        raise ValueError("No data to cache")
    else:
        raise ValueError(f"unrecognized DataFusion execution plan: {ep_str!r}")


def normalize_remote_databasetable(dt):
    return normalize_seq_with_caller(
        dt.name,
        dt.schema,
        dt.source,
        dt.namespace,
        caller="normalize_remote_databasetable",
    )


def normalize_postgres_databasetable(dt):
    from xorq.common.utils.postgres_utils import get_postgres_n_reltuples

    if dt.source.name != "postgres":
        raise ValueError(f"expected postgres backend, got {dt.source.name!r}")
    return normalize_seq_with_caller(
        dt.name,
        dt.schema,
        dt.source,
        dt.namespace,
        get_postgres_n_reltuples(dt),
        caller="normalize_postgres_databasetable",
    )


def normalize_pyiceberg_database_table(dt):
    from xorq.common.utils.pyiceberg_utils import get_iceberg_snapshots_ids

    if dt.source.name != "pyiceberg":
        raise ValueError(f"expected pyiceberg backend, got {dt.source.name!r}")

    return normalize_seq_with_caller(
        dt.name,
        dt.schema,
        dt.source,
        dt.namespace,
        get_iceberg_snapshots_ids(dt),
        caller="normalize_pyiceberg_databasetable",
    )


def normalize_snowflake_databasetable(dt):
    from xorq.common.utils.snowflake_utils import get_snowflake_last_modification_time

    if dt.source.name != "snowflake":
        raise ValueError(f"expected snowflake backend, got {dt.source.name!r}")
    return normalize_seq_with_caller(
        dt.name,
        dt.schema,
        dt.source,
        dt.namespace,
        get_snowflake_last_modification_time(dt).tobytes(),
        caller="normalize_snowflake_databasetable",
    )


def normalize_bigquery_databasetable(dt):
    # https://stackoverflow.com/questions/44288261/get-the-last-modified-date-for-all-bigquery-tables-in-a-bigquery-project/44290543#44290543
    if dt.source.name != "bigquery":
        raise ValueError(f"expected bigquery backend, got {dt.source.name!r}")
    # https://stackoverflow.com/a/44290543
    query = f"""
    SELECT last_modified_time
    FROM `{dt.namespace.database}.__TABLES__` where table_id = '{dt.name}'
    """
    ((last_modified_time,),) = dt.source.raw_sql(query).to_dataframe()
    return normalize_seq_with_caller(
        dt.name,
        dt.schema,
        dt.source,
        dt.namespace,
        last_modified_time,
    )


def normalize_duckdb_databasetable(dt):
    if dt.source.name != "duckdb":
        raise ValueError(f"expected duckdb backend, got {dt.source.name!r}")
    name = sg.table(dt.name, quoted=dt.source.compiler.quoted).sql(
        dialect=dt.source.name
    )

    ((_, plan),) = dt.source.raw_sql(f"EXPLAIN SELECT * FROM {name}").fetchall()
    scan_line = plan.split("\n")[1]
    execution_plan_name = r"\s*│\s*(\w+)\s*│\s*"
    match re.match(execution_plan_name, scan_line).group(1):
        case "ARROW_SCAN" | "PANDAS_SCAN":
            return normalize_memory_databasetable(dt)
        case "READ_PARQUET" | "READ_CSV" | "SEQ_SCAN":
            return normalize_duckdb_file_read(dt)
        case _:
            raise NotImplementedError(scan_line)


def normalize_sqlite_database_table(dt):
    from xorq.common.utils.sqlite_utils import get_sqlite_stats

    if dt.source.name != "sqlite":
        raise ValueError(f"expected sqlite backend, got {dt.source.name!r}")

    if dt.source.is_in_memory():
        return normalize_memory_databasetable(dt)
    else:
        return normalize_seq_with_caller(
            dt.name,
            dt.schema,
            dt.source,
            dt.namespace,
            get_sqlite_stats(dt),
            caller="normalize_sqlite_database_table",
        )

    pass


def normalize_duckdb_file_read(dt):
    name = sg.exp.convert(dt.name).sql(dialect=dt.source.name)
    (sql_ddl_statement,) = dt.source.con.sql(
        f"select sql from duckdb_views() where view_name = {name} UNION select sql from duckdb_tables() where table_name = {name}"
    ).fetchone()
    return normalize_seq_with_caller(
        dt.schema.to_pandas(),
        # sql_ddl_statement denotes the definition of the table, expressed as SQL DDL-statement.
        sql_ddl_statement,
    )


def rename_unbound_static(op, prefix="static-name"):
    count = itertools.count()

    def rename_unbound(node, kwargs):
        if isinstance(node, ir.UnboundTable):
            name = f"{prefix}-{next(count)}"
            return node.copy(name=name)
        else:
            if kwargs:
                return node.__recreate__(kwargs)
            return node

    return op.replace(rename_unbound)


def normalize_xorq_databasetable(dt):
    if dt.source.name != "xorq":
        raise ValueError(f"expected xorq backend, got {dt.source.name!r}")
    if isinstance(dt, rel.FlightExpr):
        return dask.base.normalize_token(
            (
                "normalize_xorq_databasetable",
                dt.input_expr,
                # we need to "stabilize" the name of the tables in the unbound expr
                rename_unbound_static(dt.unbound_expr.op()).to_expr(),
                dt.make_connection,
            )
        )
    elif isinstance(dt, rel.FlightUDXF):
        return dask.base.normalize_token(
            (
                "normalize_xorq_databasetable",
                dt.input_expr,
                dt.udxf.exchange_f,
                dt.make_connection,
            )
        )
    native_source = dt.source._sources.get_backend(dt)

    if native_source.name == "xorq":
        return normalize_datafusion_databasetable(dt)
    new_dt = rel.make_native_op(dt)
    return dask.base.normalize_token(new_dt)


@dask.base.normalize_token.register(types.ModuleType)
def normalize_module(module):
    return normalize_seq_with_caller(
        module.__name__,
        module.__package__,
        caller="normalize_module",
    )


@dask.base.normalize_token.register(dat.DataType)
def normalize_ibis_datatype(datatype):
    return normalize_seq_with_caller(datatype.name.lower(), *datatype.args)


@dask.base.normalize_token.register(rel.Read)
def normalize_read(read):
    read_kwargs = dict(read.read_kwargs)
    try_names = (
        "path",
        "paths",  # duckdb read_parquet
        "source",
        "source_list",  # duckdb
    )
    try:
        path = next(el for el in (read_kwargs.get(name) for name in try_names) if el)
    except StopIteration as err:
        raise ValueError("unable to find path name in read_kwargs") from err
    if isinstance(path, (list, tuple)):
        # normalize_filenames may have converted a single path to a list
        path = path[0] if len(path) == 1 else path
    if isinstance(path, (str, pathlib.Path)):
        path = str(path)
        if path.startswith("http") or path.startswith("https:"):
            req = urllib.request.Request(
                path, method="HEAD", headers={"User-Agent": ""}
            )
            resp = urllib.request.urlopen(req)
            headers = resp.info()
            tpls = (("url", path),) + tuple(
                (k, headers.get(k))
                for k in (
                    "Last-Modified",
                    "Content-Length",
                    "Content-Type",
                )
            )
        elif path.startswith(("s3", "gs", "gcs")):
            metadata = api.get_object_metadata(
                path, **{k: v for k, v in read_kwargs.items() if k != "path"}
            )
            tpls = tuple(
                (k, metadata.get(k))
                for k in ("location", "last_modified", "size", "e_tag", "version")
            )
        elif (path := pathlib.Path(path)).exists():
            tpls = read.normalize_method(path)
        else:
            raise NotImplementedError(f'Don\'t know how to deal with path "{path}"')
    elif isinstance(path, (list, tuple)) and all(isinstance(el, str) for el in path):
        raise NotImplementedError(f'Don\'t know how to deal with path "{path}"')
    else:
        raise NotImplementedError(f'Don\'t know how to deal with path "{path}"')
    tpls += tuple(
        (k, v)
        for k, v in read.read_kwargs
        if k
        in (
            "mode",
            "schema",
            "temporary",
        )
    )
    return normalize_seq_with_caller(
        read.schema,
        tpls,
        caller="normalize_read",
    )


@dask.base.normalize_token.register(ir.DatabaseTable)
def normalize_databasetable(dt):
    dct = {
        "pandas": normalize_pandas_databasetable,
        "datafusion": normalize_datafusion_databasetable,
        "postgres": normalize_postgres_databasetable,
        "snowflake": normalize_snowflake_databasetable,
        "xorq": normalize_xorq_databasetable,
        "duckdb": normalize_duckdb_databasetable,
        "trino": normalize_remote_databasetable,
        "gizmosql": normalize_remote_databasetable,
        "bigquery": normalize_bigquery_databasetable,
        "pyiceberg": normalize_pyiceberg_database_table,
        "sqlite": normalize_sqlite_database_table,
    }
    f = dct[dt.source.name]
    return f(dt)


@dask.base.normalize_token.register(rel.RemoteTable)
def normalize_remote_table(dt):
    if not isinstance(dt, rel.RemoteTable):
        raise ValueError(f"expected RemoteTable, got {type(dt)}")

    return normalize_seq_with_caller(
        ("schema", dt.schema),
        ("expr", dt.remote_expr),
        # only thing that matters is the type of source its going into
        ("source", dt.source.name),
        caller="normalize_remote_table",
    )


@dask.base.normalize_token.register(rel.CachedNode)
def normalize_cached_node(node):
    return normalize_seq_with_caller(
        node.parent,
        node.cache,
        caller="normalize_cached_node",
    )


@dask.base.normalize_token.register(ibis.backends.BaseBackend)
def normalize_backend(con):
    name = con.name
    if name == "snowflake":
        con_details = con.con._host
    elif name == "postgres":
        con_dct = con.con.info.get_parameters() | {"port": con.con.info.port}
        con_details = {k: con_dct[k] for k in ("host", "port", "dbname")}
    elif name == "pandas":
        con_details = id(con.dictionary)
    elif name in ("datafusion", "duckdb", "xorq", "gizmosql"):
        con_details = (con._profile.con_name, con._profile.kwargs_tuple)
    elif name == "trino":
        con_details = con.con.host
    elif name == "bigquery":
        con_details = (
            con.project_id,
            con.dataset_id,
        )
    elif name == "pyiceberg":
        catalog_params = con.catalog_params
        con_details = (con.catalog.name,) + tuple(
            catalog_params[k] for k in ("type", "uri", "warehouse")
        )
    elif name == "sqlite":
        return id(con.con) if con.is_in_memory() else con.uri
    else:
        raise ValueError(f"no normalization rule for backend {name!r}")
    return normalize_seq_with_caller(
        name,
        con_details,
    )


@dask.base.normalize_token.register(ir.Schema)
def normalize_schema(schema):
    return normalize_seq_with_caller(
        schema.to_pandas(),
    )


@dask.base.normalize_token.register(ir.Namespace)
def normalize_namespace(ns):
    return normalize_seq_with_caller(
        ns.catalog,
        ns.database,
    )


@dask.base.normalize_token.register(InputType)
def tokenize_input_type(obj):
    return normalize_seq_with_caller(
        obj.__class__.__module__, obj.__class__.__name__, obj.name, obj.value
    )


@dask.base.normalize_token.register(ScalarUDF)
def normalize_scalar_udf(udf):
    typs = tuple(arg.dtype for arg in udf.args)
    return normalize_seq_with_caller(
        ScalarUDF,
        typs,
        udf.dtype,
        udf.__func__,
        #
        # ExprScalarUDF
        udf.__config__.get("computed_kwargs_expr"),
        # we are insensitive to these for now
        # udf.__udf_namespace__,
        # udf.__func_name__,
        caller="normalize_scalar_udf",
    )


@dask.base.normalize_token.register(AggUDF)
def normalize_agg_udf(udf):
    (*args, where) = udf.args
    if where is not None:
        # TODO: determine if sql string already contains
        #       the relevant information of `where`
        raise NotImplementedError
    typs = tuple(arg.dtype for arg in args)
    return normalize_seq_with_caller(
        AggUDF,
        typs,
        udf.dtype,
        udf.__func__,
        # we are insensitive to these for now
        # udf.__udf_namespace__,
        # udf.__func_name__,
        caller="normalize_agg_udf",
    )


def opaque_node_replacer(node, kwargs):
    # FIXME: use xorq.common.utils.graph_utils.opaque_ops (includes ExprScalarUDF)
    opaque_ops = (
        rel.Read,
        rel.CachedNode,
        rel.RemoteTable,
        rel.FlightUDXF,
        rel.FlightExpr,
        # udf.ExprScalarUDF,
    )
    match node:
        case rel.CachedNode():
            new_node = api.table(
                node.schema,
                name=dask.base.tokenize(node.parent, node.cache),
            ).op()
        case rel.Read():
            new_node = api.table(
                node.schema,
                name=dask.base.tokenize(node),
            ).op()
        case rel.RemoteTable():
            new_node = api.table(
                node.schema,
                name=dask.base.tokenize(node.remote_expr),
            ).op()
        case rel.FlightUDXF() | rel.FlightExpr():
            new_node = api.table(
                node.schema,
                name=dask.base.tokenize(node),
            ).op()
        # # FIXME: what to do about ExprScalarUDF?
        # case udf.ExprScalarUDF():
        #     # ExprScalarUDF doesn't have a schema
        #     # it has computed_kwargs_expr and others
        #     node = xo.table(
        #         node.schema,
        #         name=dask.base.tokenize(node),
        #     ).op()
        case _:
            if isinstance(node, opaque_ops):
                raise ValueError(f"unhandled opaque node type: {type(node)}")
            elif kwargs:
                new_node = node.__recreate__(kwargs)
            else:
                new_node = node
    return new_node


@dask.base.normalize_token.register(ibis.expr.types.Expr)
def normalize_expr(expr):
    from xorq.expr.api import get_compiler

    return normalize_op(expr.op(), compiler=get_compiler(expr))


def normalize_op(op, compiler=None):
    sql = unbound_expr_to_default_sql(
        op.replace(opaque_node_replacer).to_expr().unbind(),
        compiler=compiler,
    )
    reads = op.find(rel.Read)
    dts = tuple(
        node
        for node in op.find(ir.DatabaseTable)
        if not isinstance(node, (rel.CachedNode, rel.Read))
    )
    udfs = op.find((AggUDF, ScalarUDF))
    mems = op.find(ir.InMemoryTable)
    token = normalize_seq_with_caller(
        sql,
        reads,
        dts,
        udfs,
        tuple(map(normalize_inmemorytable, mems)),
        caller="normalize_expr",
    )
    return token



import functools
import pathlib

from xorq.common.utils.defer_utils import (
    deferred_read_csv,
    deferred_read_parquet,
)
from xorq.config import options


whitelist = [
    "astronauts",
    "awards_players",
    "batting",
    "diamonds",
    "functional_alltypes",
    "iris",
    "penguins",
    "hn_posts_nano",
    "hn-data-small.parquet",
]


@functools.cache
def get_name_to_suffix() -> dict[str, str]:
    board = options.pins.get_board()
    dct = {
        name: pathlib.Path(board.pin_meta(name).file).suffix
        for name in board.pin_list()
        if name in whitelist
    }
    return dct


def get_table_from_name(name, backend, table_name=None, deferred=True, **kwargs):
    suffix = get_name_to_suffix().get(name)
    match suffix:
        case ".parquet":
            if deferred:
                method = functools.partial(deferred_read_parquet, con=backend)
            else:
                method = backend.read_parquet
        case ".csv":
            if deferred:
                method = functools.partial(deferred_read_csv, con=backend)
            else:
                method = backend.read_csv
        case _:
            raise ValueError(
                f"unsupported file suffix {suffix!r}; expected .parquet or .csv"
            )
    path = options.pins.get_path(name)
    return method(path, table_name=table_name or name, **kwargs)



import pytest
from pytest import param

import xorq.api as xo
from xorq.tests.util import assert_frame_equal


@pytest.fixture(scope="function")
def con():
    return xo.connect()


def test_fill_null_with_dict(compiler, t):
    expr = t.fill_null({"a": 0, "b": "unknown"})
    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]

    assert expression["op"] == "FillNull"
    assert "a" in expression["replacements"]
    assert "b" in expression["replacements"]

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        param(xo.null().fill_null(5), 5, id="na_fill_null"),
        param(xo.literal(5).fill_null(10), 5, id="non_na_fill_null"),
    ],
)
def test_scalar_fill_null_nullif(con, expr, expected, compiler):
    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)

    assert con.execute(roundtrip_expr) == expected


@pytest.mark.parametrize(
    "replacements",
    [
        param({"int_col": 20}, id="int"),
        param({"double_col": -1, "string_col": "missing"}, id="double-int-str"),
        param({"double_col": -1.5, "string_col": "missing"}, id="double-str"),
        param({}, id="empty"),
    ],
)
def test_table_fill_null_mapping(replacements, con, compiler, parquet_dir):
    alltypes = xo.deferred_read_parquet(
        parquet_dir / "functional_alltypes.parquet", con=con
    )

    table = alltypes.mutate(
        int_col=alltypes.int_col.nullif(1),
        double_col=alltypes.double_col.nullif(3.0),
        string_col=alltypes.string_col.nullif("2"),
    ).select("id", "int_col", "double_col", "string_col")

    profiles = {
        con._profile.hash_name: con,
    }

    expr = table.fill_null(replacements)
    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict, profiles)

    result = roundtrip_expr.execute().reset_index(drop=True)
    pd_table = table.execute()
    expected = pd_table.fillna(replacements).reset_index(drop=True)

    assert_frame_equal(result, expected, check_dtype=False)


def test_table_unnest(con, compiler):
    t = con.create_table("test", {"x": [[1, 2, 3], [4, 5], [6]]})

    expr = t.unnest("x")

    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]

    assert expression["op"] == "TableUnnest"
    assert "parent" in expression
    assert "column" in expression
    assert "column_name" in expression
    assert "keep_empty" in expression

    profiles = {
        con._profile.hash_name: con,
    }
    roundtrip_expr = compiler.from_yaml(yaml_dict, profiles)

    result = roundtrip_expr.execute().reset_index(drop=True)
    expected = expr.execute().reset_index(drop=True)

    assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    ("offset", "keep_empty"),
    [
        param(None, False, id="no_offset_no_keep_empty"),
        param("idx", False, id="with_offset_no_keep_empty"),
        param(None, True, id="no_offset_with_keep_empty"),
    ],
)
def test_table_unnest_options(con, compiler, offset, keep_empty):
    t = con.create_table("test", {"x": [[1, 2], [], [3]]})

    expr = t.unnest("x", offset=offset, keep_empty=keep_empty)

    yaml_dict = compiler.to_yaml(expr)

    expression = yaml_dict["expression"]
    assert expression["op"] == "TableUnnest"

    roundtrip_expr = compiler.from_yaml(
        yaml_dict,
        profiles={
            con._profile.hash_name: con,
        },
    )

    assert_frame_equal(
        con.execute(roundtrip_expr).reset_index(drop=True),
        expr.execute().reset_index(drop=True),
    )



import re


def make_name(prefix, to_tokenize):
    import dask

    from xorq.ibis_yaml.config import config

    tokenized = dask.base.tokenize(to_tokenize)
    name = ("_" + prefix + "_" + tokenized)[: config.hash_length].lower()
    return name


def _clean_udf_name(udf_name):
    if udf_name.isidentifier():
        return udf_name
    else:
        return f"fun_{re.sub(r'[^0-9a-zA-Z_]', '_', udf_name)}".lower()


def get_uid_prefix(name, pattern="^(ibis_[\\w-]+_)\\w{26}$"):
    # xorq.vendor.ibis.util.gen_name
    if match := re.match(pattern, name):
        return match.group(1)
    else:
        return None



import cProfile
import inspect
import pstats
import tempfile
import time
from contextlib import contextmanager
from pstats import SortKey


def profile(stmt, field=SortKey.CUMULATIVE):
    f_back = inspect.currentframe().f_back
    with tempfile.NamedTemporaryFile() as ntf:
        cProfile.runctx(
            stmt,
            globals=f_back.f_globals,
            locals=f_back.f_locals,
            filename=ntf.name,
        )
        p = pstats.Stats(ntf.name)
    if field is not None:
        p = p.sort_stats(field)
    return p


@contextmanager
def timed(span, logger, event_name, **extra):
    t = time.monotonic()
    yield lambda: time.monotonic() - t
    elapsed = time.monotonic() - t
    payload = {"elapsed_s": elapsed, **extra}
    span.add_event(event_name, payload)
    logger.info(event_name, **payload)



import contextlib
import functools
import itertools
import warnings
from typing import Any

import pandas as pd
import pyarrow as pa
import sqlglot as sg
import sqlglot.expressions as sge
import toolz

import xorq.vendor.ibis.expr.api as api
import xorq.vendor.ibis.expr.schema as sch
import xorq.vendor.ibis.expr.types as ir
from xorq.common.utils.logging_utils import get_logger
from xorq.expr.relations import (
    prepare_create_table_from_expr,
)
from xorq.vendor.ibis.backends.snowflake import _SNOWFLAKE_MAP_UDFS
from xorq.vendor.ibis.backends.snowflake import Backend as IbisSnowflakeBackend
from xorq.vendor.ibis.expr.operations.relations import (
    Namespace,
)


try:
    from enum import StrEnum
except ImportError:
    from strenum import StrEnum


logger = get_logger(__name__)


class SnowflakeAuthenticator(StrEnum):
    # https://docs.snowflake.com/en/developer-guide/node-js/nodejs-driver-options#label-nodejs-auth-options
    password = "none"
    mfa = "username_password_mfa"
    keypair = "snowflake_jwt"
    sso = "externalbrowser"
    # oauth = "oauth"
    # oauth2 = "oauth_authorization_code"


@functools.wraps(IbisSnowflakeBackend.do_connect)
def wrapped_do_connect(self, create_object_udfs: bool = None, **kwargs: Any):
    from xorq.common.utils.snowflake_keypair_utils import maybe_decrypt_private_key
    from xorq.common.utils.snowflake_utils import default_create_object_udfs

    if create_object_udfs is None:
        create_object_udfs = default_create_object_udfs

    if "private_key" in kwargs:
        kwargs = maybe_decrypt_private_key(kwargs)
    return IbisSnowflakeBackend.do_connect(
        self, create_object_udfs=create_object_udfs, **kwargs
    )


class Backend(IbisSnowflakeBackend):
    _top_level_methods = (
        "connect_env",
        "connect_env_mfa",
        "connect_env_password",
        "connect_env_keypair",
    )

    @staticmethod
    def connect_env(
        passcode=None,
        authenticator=None,
        **kwargs,
    ):
        from xorq.common.utils.snowflake_utils import make_connection

        return make_connection(
            authenticator=authenticator,
            passcode=passcode,
            **kwargs,
        )

    connect_env_password = staticmethod(
        toolz.curry(connect_env, authenticator=SnowflakeAuthenticator.password)
    )

    connect_env_mfa = staticmethod(
        toolz.curry(connect_env, authenticator=SnowflakeAuthenticator.mfa)
    )

    connect_env_keypair = staticmethod(
        toolz.curry(connect_env, authenticator=SnowflakeAuthenticator.keypair)
    )

    def table(self, *args, **kwargs):
        table = super().table(*args, **kwargs)
        op = table.op()
        if op.namespace == Namespace(None, None):
            (catalog, database) = (self.current_catalog, self.current_database)
            table = op.copy(**{"namespace": Namespace(catalog, database)}).to_expr()
        return table

    def create_table(
        self,
        name: str,
        obj: pd.DataFrame | pa.Table | ir.Table | None = None,
        *,
        schema: sch.Schema | None = None,
        database: str | None = None,
        temp: bool = False,
        overwrite: bool = False,
        comment: str | None = None,
    ) -> ir.Table:
        """Create a table in Snowflake.

        Parameters
        ----------
        name
            Name of the table to create
        obj
            The data with which to populate the table; optional, but at least
            one of `obj` or `schema` must be specified
        schema
            The schema of the table to create; optional, but at least one of
            `obj` or `schema` must be specified
        database
            The name of the database in which to create the table; if not
            passed, the current database is used.
        temp
            Create a temporary table
        overwrite
            If `True`, replace the table if it already exists, otherwise fail
            if the table exists
        comment
            Add a comment to the table

        """
        if obj is None and schema is None:
            raise ValueError("Either `obj` or `schema` must be specified")

        quoted = self.compiler.quoted

        if database is None:
            target = sg.table(name, quoted=quoted)
            catalog = db = database
        else:
            db = self._warn_and_create_table_loc(database=database)
            (catalog, db) = (db.catalog, db.db)
            target = sg.table(name, db=db, catalog=catalog, quoted=quoted)

        column_defs = [
            sge.ColumnDef(
                this=sg.to_identifier(name, quoted=quoted),
                kind=self.compiler.type_mapper.from_ibis(typ),
                constraints=(
                    None
                    if typ.nullable
                    else [sge.ColumnConstraint(kind=sge.NotNullColumnConstraint())]
                ),
            )
            for name, typ in (schema or {}).items()
        ]

        if column_defs:
            target = sge.Schema(this=target, expressions=column_defs)

        properties = []

        if temp:
            properties.append(sge.TemporaryProperty())

        if comment is not None:
            properties.append(sge.SchemaCommentProperty(this=sge.convert(comment)))

        if obj is not None:
            if not isinstance(obj, ir.Expr):
                table = api.memtable(obj)
            else:
                if catalog is not None and db is not None:
                    table = prepare_create_table_from_expr(
                        self, obj, database=(catalog, db)
                    )
                else:
                    table = prepare_create_table_from_expr(self, obj)

            self._run_pre_execute_hooks(table)

            query = self.compiler.to_sqlglot(table)
        else:
            query = None

        create_stmt = sge.Create(
            kind="TABLE",
            this=target,
            replace=overwrite,
            properties=sge.Properties(expressions=properties),
            expression=query,
        )

        with self._safe_raw_sql(create_stmt):
            pass

        return self.table(name, database=(catalog, db))

    do_connect = wrapped_do_connect

    def _setup_session(self, *, session_parameters, create_object_udfs: bool):
        con = self.con

        # enable multiple SQL statements by default
        session_parameters.setdefault("MULTI_STATEMENT_COUNT", 0)
        # don't format JSON output by default
        session_parameters.setdefault("JSON_INDENT", 0)

        # overwrite session parameters that are required for ibis + snowflake
        # to work
        session_parameters.update(
            {
                # Use Arrow for query results
                "PYTHON_CONNECTOR_QUERY_RESULT_FORMAT": "arrow_force",
                # JSON output must be strict for null versus undefined
                "STRICT_JSON_OUTPUT": True,
                # Timezone must be UTC
                "TIMEZONE": "UTC",
            },
        )

        with contextlib.closing(con.cursor()) as cur:
            cur.execute(
                "ALTER SESSION SET {}".format(
                    " ".join(f"{k} = {v!r}" for k, v in session_parameters.items())
                )
            )

        # snowflake activates a database on creation, so reset it back
        # to the original database and schema
        if con.database and "/" in con.database:
            (catalog, db) = con.database.split("/")
            use_stmt = sge.Use(
                kind="SCHEMA",
                this=sg.table(db, catalog=catalog, quoted=self.compiler.quoted),
            ).sql(dialect=self.name)
            with contextlib.closing(con.cursor()) as cur:
                try:
                    cur.execute(use_stmt)
                except Exception as e:  # noqa: BLE001
                    warnings.warn(f"Unable to set catalog,db: {e}", stacklevel=2)

        if create_object_udfs:
            create_stmt = sge.Create(
                kind="DATABASE", this="ibis_udfs", exists=True
            ).sql(dialect=self.name)

            stmts = [
                create_stmt,
                # snowflake activates a database on creation, so reset it back
                # to the original database and schema
                *itertools.starmap(self._make_udf, _SNOWFLAKE_MAP_UDFS.items()),
            ]

            stmt = ";\n".join(stmts)
            with contextlib.closing(con.cursor()) as cur:
                try:
                    cur.execute(stmt)
                except Exception as e:  # noqa: BLE001
                    warnings.warn(
                        f"Unable to create Ibis UDFs, some functionality will not work: {e}",
                        stacklevel=2,
                    )
        # without this self.current_{catalog,database} is not synchronized with con.{database,schema}
        with contextlib.closing(con.cursor()) as cur:
            try:
                cur.execute("SELECT CURRENT_TIME")
            except Exception:  # noqa: BLE001
                pass

    @property
    def adbc(self):
        from xorq.common.utils.snowflake_utils import SnowflakeADBC

        adbc = SnowflakeADBC(self)
        return adbc

    def read_record_batches(
        self,
        record_batches: pa.RecordBatchReader,
        table_name: str | None = None,
        temporary: bool = False,
        mode: str = "create",
        database=None,
        **kwargs: Any,
    ) -> ir.Table:
        logger.info(
            "reading record batches with SnowflakeADBC",
            **{
                "table_name": table_name,
                "temporary": temporary,
                "mode": mode,
                **kwargs,
            },
        )

        snowflake_adbc = self.adbc
        snowflake_adbc.adbc_ingest(
            table_name, record_batches, mode=mode, database=database, **kwargs
        )
        return self.table(table_name, database=database)


def connect(*args, **kwargs):
    con = Backend(*args, **kwargs)
    con.reconnect()
    return con



import dask

import xorq.common.utils.dask_normalize.dask_normalize_expr  # noqa: F401
import xorq.common.utils.dask_normalize.dask_normalize_function  # noqa: F401
import xorq.common.utils.dask_normalize.dask_normalize_other  # noqa: F401


dask.config.set({"tokenize.ensure-deterministic": True})


@dask.base.normalize_token.register(object)
def raise_generic_object(o):
    method = getattr(o, "__dask_tokenize__", None)
    if method is not None and not isinstance(o, type):
        return method()
    raise ValueError(f"Object {o!r} cannot be deterministically hashed")


def get_normalize_token_subset(value=dask.tokenize.normalize_object):
    return {k: v for k, v in dask.base.normalize_token._lookup.items() if v == value}


# we have to clear out any classes that might have already been registered to dask's default normalize_object
bad_keys = tuple(get_normalize_token_subset(dask.tokenize.normalize_object))
for bad_key in bad_keys:
    dask.base.normalize_token._lookup.pop(bad_key)



import dask
import numpy as np
import pyarrow as pa

from xorq.common.utils.dask_normalize.dask_normalize_utils import (
    normalize_seq_with_caller,
)


def lazy_register_pandas():
    import pandas as pd

    @dask.base.normalize_token.register(pd._libs.interval.Interval)
    def normalize_interval(interval):
        objs = (interval.left, interval.right, interval.closed)
        return normalize_seq_with_caller(*objs)

    @dask.base.normalize_token.register(pd._libs.tslibs.timestamps.Timestamp)
    def normalize_timestamp(timestamp):
        objs = (str(timestamp),)
        return normalize_seq_with_caller(*objs)


def safe_lazy_register(toplevel, function):
    if existing := dask.base.normalize_token._lazy.get(toplevel):

        def do_both():
            existing()
            function()

        to_register = do_both
    else:
        to_register = function
    dask.base.normalize_token.register_lazy(toplevel, to_register)


safe_lazy_register("pandas", lazy_register_pandas)


def lazy_register_sklearn():
    from sklearn.base import BaseEstimator

    from xorq.ibis_yaml.utils import freeze

    @dask.base.normalize_token.register(BaseEstimator)
    def normalize_sklearn_estimator(estimator):
        params = freeze(tuple(estimator.get_params(deep=True).items()))
        return normalize_seq_with_caller(
            type(estimator).__name__,
            type(estimator).__module__,
            params,
            caller="normalize_sklearn_estimator",
        )


safe_lazy_register("sklearn", lazy_register_sklearn)


@dask.base.normalize_token.register(dict)
def normalize_dict(dct):
    return normalize_seq_with_caller(*sorted(dct.items()))


@dask.base.normalize_token.register(np.random.RandomState)
def normalize_random_state(random_state):
    return normalize_seq_with_caller(random_state.get_state())


@dask.base.normalize_token.register(type(np.mean))
def normalize_numpy_array_function_dispatcher(func):
    """Normalize numpy array function dispatchers (e.g., np.mean, np.sum).

    Used by sklearn estimators like FeatureAgglomeration that have pooling_func=np.mean.
    """
    return normalize_seq_with_caller(
        func.__name__,
        func.__module__,
        caller="normalize_numpy_array_function_dispatcher",
    )


@dask.base.normalize_token.register(type)
def normalize_type(typ):
    return normalize_seq_with_caller(typ.__name__, typ.__module__)


@dask.base.normalize_token.register(pa.Table)
def normalize_pyarrow_table(table: pa.Table):
    return normalize_seq_with_caller(
        tuple(
            dask.base.tokenize(el.serialize().to_pybytes()) for el in table.to_batches()
        ),
        caller="normalize_pyarrow_table",
    )


@dask.base.normalize_token.register(pa.Schema)
def normalize_pyarrow_schema(schema: pa.Schema):
    from xorq.vendor.ibis import Schema

    return normalize_seq_with_caller(
        Schema.from_pyarrow(schema).to_pandas(),
        caller="normalize_pyarrow_schema",
    )



"""Ibis utility functions."""

from __future__ import annotations

import base64
import collections
import collections.abc
import functools
import glob
import importlib.metadata
import itertools
import operator
import os
import re
import sys
import textwrap
import types
import uuid
import warnings
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, Any, Generic, TypeVar
from uuid import uuid4

import toolz

from xorq.vendor.ibis.common.typing import Coercible


if TYPE_CHECKING:
    from collections.abc import Callable, Iterator, Sequence
    from numbers import Real

    import xorq.vendor.ibis.expr.types as ir

T = TypeVar("T", covariant=True)
S = TypeVar("S", bound=T, covariant=True)
U = TypeVar("U", covariant=True)
K = TypeVar("K")
V = TypeVar("V")


# https://www.compart.com/en/unicode/U+22EE
VERTICAL_ELLIPSIS = "\u22ee"
# https://www.compart.com/en/unicode/U+2026
HORIZONTAL_ELLIPSIS = "\u2026"


def guid() -> str:
    """Return a uuid4 hexadecimal value."""
    return uuid4().hex


def indent(text: str, spaces: int) -> str:
    """Apply an indentation using the given spaces into the given text.

    Parameters
    ----------
    text
        Text to indent
    spaces
        Number of leading spaces per line

    Returns
    -------
    str
        Indented text

    """
    prefix = " " * spaces
    return textwrap.indent(text, prefix=prefix)


def is_one_of(values: Sequence[T], t: type[U]) -> Iterator[bool]:
    """Check if the type of each value is the same of the given type.

    Parameters
    ----------
    values
        Input values
    t
        Type to check against

    Returns
    -------
    tuple

    """
    return (isinstance(x, t) for x in values)


any_of = toolz.compose(any, is_one_of)
all_of = toolz.compose(all, is_one_of)


def promote_list(val: V | Sequence[V]) -> list[V]:
    """Ensure that the value is a list.

    Parameters
    ----------
    val
        Value to promote

    Returns
    -------
    list

    """
    if isinstance(val, list):
        return val
    elif isinstance(val, dict):
        return [val]
    elif is_iterable(val):
        return list(val)
    elif val is None:
        return []
    else:
        return [val]


def promote_tuple(val: V | Sequence[V]) -> tuple[V]:
    """Ensure that the value is a tuple.

    Parameters
    ----------
    val
        Value to promote

    Returns
    -------
    tuple

    """
    if isinstance(val, tuple):
        return val
    elif is_iterable(val):
        return tuple(val)
    elif val is None:
        return ()
    else:
        return (val,)


def is_function(v: Any) -> bool:
    """Check if the given object is a function.

    Returns
    -------
    bool
        Whether `v` is a function

    """
    return isinstance(v, (types.FunctionType, types.LambdaType))


def log(msg: str) -> None:
    """Log `msg` using `options.verbose_log` if set, otherwise `print`."""
    from xorq.vendor.ibis.config import options

    if options.verbose:
        (options.verbose_log or print)(msg)


def approx_equal(a: Real, b: Real, eps: Real):
    """Return whether the difference between `a` and `b` is less than `eps`.

    Raises
    ------
    AssertionError

    """
    assert abs(a - b) < eps


def safe_index(elements: Sequence[int], value: int) -> int:
    """Find the location of `value` in `elements`.

    Return -1 if `value` is not found instead of raising `ValueError`.

    Parameters
    ----------
    elements
        Elements to index into
    value : int
        Index of the given sequence/elements

    Returns
    -------
    int

    Examples
    --------
    >>> sequence = [1, 2, 3]
    >>> safe_index(sequence, 2)
    1
    >>> safe_index(sequence, 4)
    -1

    """
    try:
        return elements.index(value)
    except ValueError:
        return -1


def is_iterable(o: Any) -> bool:
    """Return whether `o` is iterable and not a :class:`str` or :class:`bytes`.

    Parameters
    ----------
    o : object
        Any python object

    Returns
    -------
    bool

    Examples
    --------
    >>> is_iterable("1")
    False
    >>> is_iterable(b"1")
    False
    >>> is_iterable(iter("1"))
    True
    >>> is_iterable(i for i in range(1))
    True
    >>> is_iterable(1)
    False
    >>> is_iterable([])
    True

    """
    if isinstance(o, (str, bytes)):
        return False

    try:
        iter(o)
    except TypeError:
        return False
    else:
        return True


def convert_unit(value, unit, to, floor: bool = True):
    """Convert a value between different units.

    Convert `value`, is assumed to be in units of `unit`, to units of `to`.
    If `floor` is true, then use floor division on `value` if necessary.

    Parameters
    ----------
    value
        Number or numeric ibis expression
    unit
        Unit of `value`
    to
        Unit to convert to
    floor
        Whether or not to use floor division on `value` if necessary.

    Returns
    -------
    Union[numbers.Integral, ibis.expr.types.NumericValue]
        Integer converted unit

    Examples
    --------
    >>> one_second = 1000
    >>> x = convert_unit(one_second, "ms", "s")
    >>> x
    1
    >>> one_second = 1
    >>> x = convert_unit(one_second, "s", "ms")
    >>> x
    1000
    >>> x = convert_unit(one_second, "s", "s")
    >>> x
    1
    >>> x = convert_unit(one_second, "s", "M")
    Traceback (most recent call last):
        ...
    ValueError: Cannot convert to or from unit ... to unit ...

    """
    # Don't do anything if from and to units are equivalent
    if unit == to:
        return value

    units = ("W", "D", "h", "m", "s", "ms", "us", "ns")
    factors = (7, 24, 60, 60, 1000, 1000, 1000)

    monthly_units = ("Y", "Q", "M")
    monthly_factors = (4, 3)

    try:
        i, j = units.index(unit), units.index(to)
    except ValueError:
        try:
            i, j = monthly_units.index(unit), monthly_units.index(to)
            factors = monthly_factors
        except ValueError:
            raise ValueError(
                f"Cannot convert interval value from unit {unit} to unit {to}"
            )

    factor = functools.reduce(operator.mul, factors[min(i, j) : max(i, j)], 1)
    assert factor > 1

    if i < j:
        op = operator.mul
    else:
        assert i > j
        op = operator.floordiv if floor else operator.truediv
    try:
        return op(value.to_expr(), factor).op()
    except AttributeError:
        return op(value, factor)


# taken from the itertools documentation
def consume(iterator: Iterator[T], n: int | None = None) -> None:
    """Advance `iterator` n-steps ahead. If `n` is `None`, consume entirely."""
    # Use functions that consume iterators at C speed.
    if n is None:
        # feed the entire iterator into a zero-length deque
        collections.deque(iterator, maxlen=0)
    else:
        # advance to the empty slice starting at position n
        next(itertools.islice(iterator, n, n), None)


def flatten_iterable(iterable):
    """Recursively flatten the iterable `iterable`."""
    if not is_iterable(iterable):
        raise TypeError("flatten is only defined for non-str iterables")

    for item in iterable:
        if is_iterable(item):
            yield from flatten_iterable(item)
        else:
            yield item


def deprecated_msg(name, *, instead, as_of="", removed_in=""):
    msg = f"`{name}` is deprecated"

    msgs = []

    if as_of:
        msgs.append(f"as of v{as_of}")

    if removed_in:
        msgs.append(f"removed in v{removed_in}")

    if msgs:
        msg += f" {', '.join(msgs)}"
    msg += f"; {instead}"
    return msg


def warn_deprecated(name, *, instead, as_of="", removed_in="", stacklevel=1):
    """Warn about deprecated usage.

    The message includes a stacktrace and what to do instead.
    """

    msg = deprecated_msg(name, instead=instead, as_of=as_of, removed_in=removed_in)
    warnings.warn(msg, FutureWarning, stacklevel=stacklevel + 1)


def append_admonition(
    func: Callable, *, msg: str, body: str = "", kind: str = "warning"
) -> str:
    """Append a `kind` admonition with `msg` to `func`'s docstring."""
    if docstr := func.__doc__:
        preamble, *rest = docstr.split("\n\n", maxsplit=1)

        # count leading spaces and add them to the deprecation warning so the
        # docstring parses correctly
        leading_spaces = " " * sum(
            1 for _ in itertools.takewhile(str.isspace, rest[0] if rest else [])
        )

        lines = [f"::: {{.callout-{kind}}}", f"## {msg}", ":::"]
        admonition_doc = textwrap.indent("\n".join(lines), leading_spaces)

        if body:
            rest = [indent(body, spaces=len(leading_spaces) + 4), *rest]

        docstr = "\n\n".join([preamble, admonition_doc, *rest])
    else:
        lines = [f"::: {{.callout-{kind}}}", f"## {msg}", ":::"]
        admonition_doc = "\n".join(lines)
        if body:
            admonition_doc += f"\n\n{indent(body, spaces=4)}"
        docstr = admonition_doc
    return docstr


def deprecated(*, instead: str, as_of: str = "", removed_in: str = ""):
    """Decorate to warn of deprecated usage and what to do instead."""

    def decorator(func):
        msg = deprecated_msg(
            func.__qualname__, instead=instead, as_of=as_of, removed_in=removed_in
        )

        func.__doc__ = append_admonition(func, msg=f"DEPRECATED: {msg}")

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            warn_deprecated(
                func.__qualname__,
                instead=instead,
                as_of=as_of,
                removed_in=removed_in,
                stacklevel=2,
            )
            return func(*args, **kwargs)

        return wrapper

    return decorator


def backend_sensitive(
    *,
    msg: str = "This operation differs between backends.",
    why: str = "",
):
    """Indicate that an API may be sensitive to a backend."""

    def wrapper(func):
        func.__doc__ = append_admonition(func, msg=msg, body=why, kind="info")
        return func

    return wrapper


def experimental(func):
    """Decorate a callable to add warning about API instability in docstring."""

    func.__doc__ = append_admonition(
        func, msg="This API is experimental and subject to change."
    )
    return func


def backend_entry_points() -> list[importlib.metadata.EntryPoint]:
    """Get the list of installed `ibis.backend` entrypoints."""

    eps = importlib.metadata.entry_points(group="ibis.backends")
    return sorted(eps)


_common_package_aliases = {
    "pa": "pyarrow",
    "pd": "pandas",
    "np": "numpy",
    "sk": "sklearn",
    "sp": "scipy",
    "tf": "tensorflow",
}


def unalias_package(name: str) -> str:
    return _common_package_aliases.get(name, name)


def import_object(qualname: str) -> Any:
    """Attempt to import an object given its full qualname.

    Examples
    --------
    >>> ex = import_object("ibis.examples")

    Is the same as

    >>> from xorq.vendor.ibis import examples as ex

    """
    mod_name, name = qualname.rsplit(".", 1)
    mod = importlib.import_module(mod_name)
    try:
        return getattr(mod, name)
    except AttributeError:
        raise ImportError(f"cannot import name {name!r} from {mod_name!r}") from None


def normalize_filename(source: str | Path) -> str:
    source = str(source)
    for prefix in (
        "parquet",
        "csv",
        "csv.gz",
        "txt",
        "txt.gz",
        "tsv",
        "tsv.gz",
        "file",
    ):
        source = source.removeprefix(f"{prefix}://")

    def _absolufy_paths(name):
        if re.search(r"^(?:.+)://", name) is None:
            return os.path.abspath(name)
        return name

    source = _absolufy_paths(source)
    return source


def normalize_filenames(source_list):
    def _maybe_glob_pattern(maybe_pattern):
        return (
            not isinstance(maybe_pattern, Path)
            and not os.path.exists(maybe_pattern)
            and re.search(r"^(?:.+)://", maybe_pattern) is None
        )

    # Promote to list
    source_list = promote_list(source_list)

    paths = []
    for path in source_list:
        if _maybe_glob_pattern(path):
            paths.extend(glob.glob(path))
        else:
            paths.append(path)

    paths = list(map(normalize_filename, paths))

    if not paths:
        raise ValueError("At least one path is required")

    return paths


def gen_name(namespace: str) -> str:
    """Create a unique identifier."""
    uid = base64.b32encode(uuid.uuid4().bytes).decode().rstrip("=").lower()
    return f"ibis_{namespace}_{uid}"


def slice_to_limit_offset(
    what: slice, count: ir.IntegerScalar
) -> tuple[int | ir.IntegerScalar, int | ir.IntegerScalar]:
    """Convert a Python [`slice`](slice) to a `limit`, `offset` pair.

    Parameters
    ----------
    what
        The slice to convert
    count
        The total number of rows in the table as an expression

    Returns
    -------
    tuple[int | ir.IntegerScalar, int | ir.IntegerScalar]
        The offset and limit to use in a `Table.limit` call

    Examples
    --------
    >>> import ibis
    >>> t = ibis.table(dict(a="int", b="string"), name="t")

    First 10 rows
    >>> count = t.count()
    >>> what = slice(0, 10)
    >>> limit, offset = slice_to_limit_offset(what, count)
    >>> limit
    10
    >>> offset
    0

    Last 10 rows
    >>> what = slice(-10, None)
    >>> limit, offset = slice_to_limit_offset(what, count)
    >>> limit
    10
    >>> offset
    r0 := UnboundTable: t
      a int64
      b string
    <BLANKLINE>
    Add(CountStar(t), -10): CountStar(r0) + -10

    From 5th row to 10th row
    >>> what = slice(5, 10)
    >>> limit, offset = slice_to_limit_offset(what, count)
    >>> limit, offset
    (5, 5)

    """
    if (step := what.step) is not None and step != 1:
        raise ValueError("Slice step can only be 1")

    from xorq.vendor import ibis

    start = what.start
    stop = what.stop

    if start is None or start >= 0:
        offset = start or 0

        if stop is None:
            limit = None
        elif stop == 0:
            limit = 0
        elif stop < 0:
            limit = count + (stop - offset)
        else:  # stop > 0
            limit = max(stop - offset, 0)
    else:  # start < 0
        offset = count + start

        if stop is None:
            limit = -start
        elif stop == 0:
            limit = offset = 0
        elif stop < 0:
            limit = max(stop - start, 0)
            if limit == 0:
                offset = 0
        else:  # stop > 0
            limit = ibis.greatest((stop - start) - count, 0)
    return limit, offset


class Namespace:
    """Convenience class for creating patterns for various types from a module.

    Useful to reduce boilerplate when creating patterns for various types from
    a module.

    Parameters
    ----------
    factory
        The pattern to construct with the looked up types.
    module
        The module object or name to look up the types.

    """

    __slots__ = ("_factory", "_module")
    _factory: Callable
    _module: ModuleType

    def __init__(self, factory, module):
        if isinstance(module, str):
            module = sys.modules[module]
        self._module = module
        self._factory = factory

    def __getattr__(self, name: str):
        obj = getattr(self._module, name)
        return self._factory(obj)


class PseudoHashable(Coercible, Generic[V]):
    """A wrapper that provides a best effort precomputed hash."""

    __slots__ = ("obj", "hash")
    obj: V

    def __init__(self, obj: V):
        if isinstance(obj, collections.abc.Hashable):
            raise TypeError(f"Cannot wrap a hashable object: {obj!r}")
        elif isinstance(obj, collections.abc.Sequence):
            hashable_obj = tuple(obj)
        elif isinstance(obj, collections.abc.Mapping):
            hashable_obj = tuple(obj.items())
        elif isinstance(obj, collections.abc.Set):
            hashable_obj = frozenset(obj)
        else:
            hashable_obj = id(obj)

        self.obj = obj
        self.hash = hash((type(obj), hashable_obj))

    @classmethod
    def __coerce__(cls, value: V) -> PseudoHashable[V]:
        if isinstance(value, cls):
            return value
        return cls(value)

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        if isinstance(other, PseudoHashable):
            return self.obj == other.obj
        else:
            return NotImplemented

    def __ne__(self, other):
        if isinstance(other, PseudoHashable):
            return self.obj != other.obj
        else:
            return NotImplemented


def chunks(n: int, *, chunk_size: int) -> Iterator[tuple[int, int]]:
    """Return an iterator of chunk start and end indices.

    Parameters
    ----------
    n
        The total number of elements.
    chunk_size
        The size of each chunk.

    Returns
    -------
    int
        THE start and end indices of each chunk.

    Examples
    --------
    >>> list(chunks(10, chunk_size=3))
    [(0, 3), (3, 6), (6, 9), (9, 10)]
    >>> list(chunks(10, chunk_size=4))
    [(0, 4), (4, 8), (8, 10)]
    """
    return ((start, min(start + chunk_size, n)) for start in range(0, n, chunk_size))


def get_subclasses(obj: type[T]) -> Iterator[type[S]]:
    """Recursively compute all subclasses of `obj`.

    ::: {.callout-note}
    ## The resulting iterator does **not** include the input type object.
    :::

    Parameters
    ----------
    obj
        Any type object

    Examples
    --------
    >>> class Base: ...
    >>> class Subclass1(Base): ...
    >>> class Subclass2(Base): ...
    >>> class TransitiveSubclass(Subclass2): ...

    Everything inherits `Base` (directly or transitively)

    >>> list(get_subclasses(Base))
    [<class 'ibis.util.Subclass1'>, <class 'ibis.util.Subclass2'>, <class 'ibis.util.TransitiveSubclass'>]

    Nothing inherits from `Subclass1`

    >>> list(get_subclasses(Subclass1))
    []

    Only `TransitiveSubclass` inherits from `Subclass2`

    >>> list(get_subclasses(Subclass2))
    [<class 'ibis.util.TransitiveSubclass'>]

    Nothing inherits from `TransitiveSubclass`

    >>> list(get_subclasses(TransitiveSubclass))
    []
    """
    for child_class in obj.__subclasses__():
        yield child_class
        yield from get_subclasses(child_class)



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from public import public

import xorq.common.exceptions as com
import xorq.vendor.ibis.expr.operations as ops
import xorq.vendor.ibis.expr.types as ir
from xorq.vendor.ibis.common.collections import FrozenOrderedDict  # noqa: TCH001
from xorq.vendor.ibis.common.grounds import Concrete
from xorq.vendor.ibis.expr.operations.relations import Unaliased  # noqa: TCH001
from xorq.vendor.ibis.expr.types.relations import unwrap_aliases


if TYPE_CHECKING:
    from collections.abc import Sequence


@public
class WindowedTable(Concrete):
    """An intermediate table expression to hold windowing information."""

    parent: ir.Table
    time_col: ops.Column
    window_type: Literal["tumble", "hop"] | None = None
    window_size: ir.IntervalScalar | None = None
    window_slide: ir.IntervalScalar | None = None
    window_offset: ir.IntervalScalar | None = None
    groups: FrozenOrderedDict[str, Unaliased[ops.Column]] | None = None
    metrics: FrozenOrderedDict[str, Unaliased[ops.Column]] | None = None

    def __init__(self, time_col: ops.Column, **kwargs):
        if time_col is None:
            raise com.XorqInputError(
                "Window aggregations require `time_col` as an argument"
            )
        super().__init__(time_col=time_col, **kwargs)

    def tumble(
        self,
        size: ir.IntervalScalar,
        offset: ir.IntervalScalar | None = None,
    ) -> WindowedTable:
        return self.copy(window_type="tumble", window_size=size, window_offset=offset)

    def hop(
        self,
        size: ir.IntervalScalar,
        slide: ir.IntervalScalar,
        offset: ir.IntervalScalar | None = None,
    ) -> WindowedTable:
        return self.copy(
            window_type="hop",
            window_size=size,
            window_slide=slide,
            window_offset=offset,
        )

    def aggregate(
        self,
        metrics: Sequence[ir.Scalar] | None = (),
        by: str | ir.Value | Sequence[str] | Sequence[ir.Value] | None = (),
        **kwargs: ir.Value,
    ) -> ir.Table:
        by = self.parent.bind(by)
        metrics = self.parent.bind(metrics, **kwargs)

        by = unwrap_aliases(by)
        metrics = unwrap_aliases(metrics)

        groups = dict(self.groups) if self.groups is not None else {}
        groups.update(by)

        return ops.WindowAggregate(
            self.parent,
            self.window_type,
            self.time_col,
            groups=groups,
            metrics=metrics,
            window_size=self.window_size,
            window_slide=self.window_slide,
            window_offset=self.window_offset,
        ).to_expr()

    agg = aggregate

    def group_by(
        self, *by: str | ir.Value | Sequence[str] | Sequence[ir.Value]
    ) -> WindowedTable:
        by = tuple(v for v in by if v is not None)
        groups = self.parent.bind(*by)
        groups = unwrap_aliases(groups)
        return self.copy(groups=groups)



from pathlib import Path

import pyarrow as pa
import pytest

import xorq.api as xo
from xorq.conftest import array_types_df, win


expected_tables = (
    "array_types",
    "astronauts",
    "awards_players",
    "awards_players_special_types",
    "batting",
    "diamonds",
    "functional_alltypes",
    "geo",
    "geography_columns",
    "geometry_columns",
    "json_t",
    "map",
    "spatial_ref_sys",
    "topk",
    "tzone",
)


def remove_unexpected_tables(dirty):
    # drop tables
    for table in dirty.list_tables():
        if table not in expected_tables:
            dirty.drop_table(table, force=True)

    # drop view
    for table in dirty.list_tables():
        if table not in expected_tables:
            dirty.drop_view(table, force=True)

    if sorted(dirty.list_tables()) != sorted(expected_tables):
        raise ValueError(
            f"unexpected tables after cleanup: {set(dirty.list_tables()) - set(expected_tables)}"
        )


@pytest.fixture(scope="session")
def ddl_file(root_dir):
    ddl_dir = root_dir / "db" / "datafusion.sql"
    return ddl_dir


def statements(ddl_file: Path):
    return (
        statement
        for statement in map(str.strip, ddl_file.read_text().split(";"))
        if statement
    )


@pytest.fixture(scope="session")
def con(data_dir, ddl_file):
    conn = xo.connect()
    parquet_dir = data_dir / "parquet"
    conn.read_parquet(
        parquet_dir / "functional_alltypes.parquet", "functional_alltypes"
    )
    conn.read_parquet(parquet_dir / "batting.parquet", "batting")
    conn.read_parquet(parquet_dir / "diamonds.parquet", "diamonds")
    conn.read_parquet(parquet_dir / "astronauts.parquet", "astronauts")
    conn.read_parquet(parquet_dir / "awards_players.parquet", "awards_players")

    conn.create_table("array_types", array_types_df)
    conn.create_table("win", win)

    if ddl_file.is_file() and ddl_file.name.endswith(".sql"):
        for statement in statements(ddl_file):
            with conn._safe_raw_sql(statement):  # noqa
                pass

    return conn


@pytest.fixture(scope="session")
def dirty_ls_con():
    con = xo.connect()
    return con


@pytest.fixture(scope="function")
def ls_con(dirty_ls_con):
    # since we don't register, maybe just create a fresh con
    yield dirty_ls_con
    # drop tables
    for table_name in dirty_ls_con.list_tables():
        dirty_ls_con.drop_table(table_name, force=True)
    # drop view
    for table_name in dirty_ls_con.list_tables():
        dirty_ls_con.drop_view(table_name, force=True)


@pytest.fixture(scope="session")
def alltypes(con):
    return con.table("functional_alltypes")


@pytest.fixture(scope="session")
def batting(con):
    return con.table("batting")


@pytest.fixture(scope="session")
def awards_players(con):
    return con.table("awards_players")


@pytest.fixture(scope="session")
def diamonds(con):
    return con.table("diamonds")


@pytest.fixture(scope="session")
def alltypes_df(alltypes):
    return alltypes.execute()


@pytest.fixture(scope="session")
def batting_df(batting):
    return batting.execute()


@pytest.fixture(scope="session")
def sorted_df(alltypes_df):
    return alltypes_df.sort_values("id").reset_index(drop=True)


@pytest.fixture(scope="session")
def array_types(con):
    return con.table("array_types")


@pytest.fixture(scope="session")
def struct(con):
    return con.table("structs")


@pytest.fixture(scope="session")
def struct_df(struct):
    return struct.execute()


@pytest.fixture
def df():
    # create a RecordBatch and a new DataFrame from it
    batch = pa.RecordBatch.from_arrays(
        [
            pa.array([0, 1, 2, 3, 4, 5, 6]),
            pa.array([7, 4, 3, 8, 9, 1, 6]),
            pa.array(["A", "A", "A", "A", "B", "B", "B"]),
        ],
        names=["a", "b", "c"],
    )

    return batch.to_pandas()


@pytest.fixture(scope="session")
def win_table(con):
    return con.table("win")



import numpy as np
import pandas as pd
from toolz import identity

import xorq.api as xo
import xorq.expr.datatypes as dt
from xorq import udf
from xorq.common.utils.toolz_utils import curry
from xorq.vendor import ibis


df = pd.DataFrame(
    [np.int64(1), 2, 3, 4, 5, 6, None, None, 1, None, 2, None, 3, 4, None, None, 5],
    columns=["x"],
)

con = xo.connect()
data = con.register(df, table_name="data").select("x")
schema = data.schema()


@udf.agg.pandas_df(schema=schema, return_type=dt.float64, name="explode_sum")
def explode_sum(df):
    return df["x"].dropna().astype(float).sum()


expr = data.tag("full")
expr = explode_sum.on_expr(expr).name("explode_sum").as_table()


@curry
def add_value(value, frame, **kwargs):
    return (frame["x"] + value).astype(float)


add_sum = udf.make_pandas_expr_udf(
    computed_kwargs_expr=expr,
    fn=add_value,
    schema=ibis.schema({"x": dt.float64}),
    name="add_explode_sum",
    return_type=dt.float64,
    post_process_fn=identity,
)

expr = data.mutate(out=add_sum.on_expr(data)).as_table()


if __name__ == "__pytest_main__":
    assert (res := expr.execute()) is not None
    print(res)
